import tkinter as tk
from typing import Any, Callable
import json
import csv

import numpy as np

from pstl.gui import Widgets

from pstl.diagnostics.probes.langmuir.single.helpers import available_plasma_properties

def print_not_yet_implemented():
    print("Not yet implemented")


class SingleProbeLangmuirResultsFrame(tk.Frame):
    def __init__(
            self, 
            master  =   None, 
            cnf     :   dict            =   {}, 
            *args, 
            results :   dict | None     =   None, 
            fname   :   str | None      =   None, 
            ftype   :   str | None      =   None,
            **kwargs):
        
        # pop out ResultOutput kwargs out of kwargs to prevent initialize error in tk.Frame
        result_output_kwargs = kwargs.pop("results_output_kwargs",{})
        btn_kwargs = result_output_kwargs.pop("btn_kwargs",{})

        # initialize tk.Frame
        super().__init__(master, cnf, *args, **kwargs)

        # add widgets organizer
        self.widgets = Widgets()

        # if results are not given
        if results is None:
            results = {}
            for key in available_plasma_properties:
                results[key] = {"value": 1, "other": None}
        elif isinstance(results, dict):
            pass
        else:
            raise ValueError(
                "'results' must be dictionary: {}".format(type(results)))

        # create and save results as an attribute
        self._results = results

        # number of properties to display and grid positions
        grid_positions_list = self.grid_positions(results.keys())

        # loop and create labels to display
        result_frames = {}
        row = 0 
        for k, key in enumerate(results.keys()):
            # Results displayed value on label
            text = self.results[key]["value"]
            # Results units displayed
            unit = self.results[key]["other"]
            unit = self.default_unit_format(key) if unit is None else unit
            # get command for each button based on kwargs["result_output_kwargs"]["btn_kwargs"][key]
            command = btn_kwargs.get(key, print_not_yet_implemented)
            result_output_kwargs["btn_kwargs"] = {"command": command}
            # create One frame to house btn, text result, and unit display
            result_frames[key] = ResultOutput(
                key, text, unit, self, cnf, 
                **result_output_kwargs) # these are frames of btn and label
            # get grid positions 
            row = grid_positions_list[k][0]
            col = grid_positions_list[k][1]
            # pack it away!
            result_frames[key].grid(
                row=row, column=col,
                sticky="NWSE", padx=5, pady=5)
            
        # add export frame (button&entry) to save results
        frame_save = SaveResults(
            self.export_results, 
            fname=fname, ftype=ftype,
            master=self, cnf=cnf,
            *args,**kwargs)
        # pack save away!
        frame_save.grid(row=row+1,column=0,columnspan=3,sticky="NSWE", padx=5,pady=5)

        # save Individual ResultOuput Frame access
        self.widgets.frames["results"] = result_frames
        
        # save SaveResult Frame access
        self.widgets.frames["save"] = frame_save

        # function to update the results
        self.update_texts(results=self.results)

        # Configure the columns to have the same weight/size
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=1)
    
    def save(self):
        self.widgets.frames["save"].save()

    @property
    def results(self):
        return self._results

    @results.setter
    def results(self, val):
        self.update_texts(val)
        self._results = val

    def update_label(self, key, text):
        if key in self.widgets.frames["results"]:
            self.widgets.frames["results"][key].update_text(text=text)
        else:
            table = "\n".join(
                [f"{k}\t{v}" for k, v in enumerate(self.widgets.frames["results"])])
            raise ValueError(
                f"Matching key not found: {key}\nChoose from one of the available options:\n{table}")

    def update_texts(self, results=None):
        if results is None:
            results = self.results

        for key in results.keys():
            text = self.default_text_format(key)
            val = results[key]['value']
            val = 0.0 if val is None else val
            text = text.format(val)
            self.update_label(key, text)

    def export_results(self):
        temp = self.results
        results = {}

        for key in temp.keys():
            value = temp[key]["value"]
            unit = temp[key].get("unit",self.default_unit_string(key))
            results[key] = {"value":value,"unit":unit}
        return results

    def default_text_format(self, key):
        if key == "V_f":
            text = "{0:.2f}"
        elif key == "V_s":
            text = "{0:.2f}"
        elif key == "I_is":
            text = "{0:.2e}"
        elif key == "n_i":
            text = "{0:.2e}"
        elif key == "I_es":
            text = "{0:.2e}"
        elif key == "KT_e":
            text = "{0:.2f}"
        elif key == "n_e":
            text = "{0:.2e}"
        elif key == "J_es":
            text = "{0:.2e}"
        elif key == "J_is":
            text = "{0:.2e}"
        elif key == "lambda_De":
            text = "{0:.2e}"
        elif key == "r_p/lambda_De":
            text = "{0:.2e}"
        elif key == "sheath":
            text = "{0}"
        else:
            text = "{}"

        return text
    
    def default_unit_string(self, key):
        if key == "V_f":
            text = "V"
        elif key == "V_s":
            text = "V"
        elif key == "I_is":
            text = "A"
        elif key == "n_i":
            text = "m^-3"
        elif key == "I_es":
            text = "A"
        elif key == "KT_e":
            text = "eV"
        elif key == "n_e":
            text = "m^-3"
        elif key == "J_es":
            text = "A/m^2"
        elif key == "J_is":
            text = "A/m^2"
        elif key == "lambda_De":
            text = "m"
        elif key == "r_p/lambda_De":
            text = "-"
        elif key == "sheath":
            text = "-"
        else:
            text = "{}"

        return text

    def default_unit_format(self, key):
        if key == "V_f":
            text = "V"
        elif key == "V_s":
            text = "V"
        elif key == "I_is":
            text = "A"
        elif key == "n_i":
            text = "m\u207B\u00B3"
        elif key == "I_es":
            text = "A"
        elif key == "KT_e":
            text = "eV"
        elif key == "n_e":
            text = "m\u207B\u00B3"
        elif key == "J_es":
            text = "A/m\u00B2"
        elif key == "J_is":
            text = "A/m\u00B2"
        elif key == "lambda_De":
            text = "m"
        elif key == "r_p/lambda_De":
            text = ""
        elif key == "sheath":
            text = ""
        else:
            text = "N/A"

        return text


    def grid_positions(self, lst, layout="two"):
        # layout = "square"
        if layout == "two":
            func = self._two_layout
        elif layout == "square":
            func = self._square_layout
        else:
            raise ValueError("layout is not knonw: {layout}")
        return func(lst)

    def _sort_positions(self, num_items, num_rows, num_cols):
        positions = []
        for i in range(num_items):
            row = int(i/num_cols) 
            col = i % num_cols
            positions.append((row, col))

        return positions

    def _two_layout(self, lst):
        num_items = len(lst)
        num_rows = int(np.ceil(num_items/2))
        num_cols = 2
        return self._sort_positions(num_items, num_rows, num_cols)

    def _square_layout(self, lst):
        # Calculate the number of rows and columns needed for the closest square grid
        num_items = len(lst)
        grid_size = int(np.ceil(np.sqrt(num_items)))
        num_rows = grid_size
        num_cols = grid_size

        # If the grid is larger than necessary, reduce the number of columns
        if num_items < num_rows*num_cols:
            num_cols = int(np.ceil(num_items/num_rows))

        return self._sort_positions(num_items, num_rows, num_cols)

class SaveResults(tk.Frame):
    def __init__(self, command, fname=None, ftype=None, master=None, cnf={}, *args,**kwargs):
        super().__init__(master,cnf, *args, **kwargs)

        # create widgets
        self.widgets = Widgets()

        # create save button
        btn_kwargs = kwargs.get("btn_kwargs", {})
        btn_kwargs.setdefault("text", "Save")
        btn_kwargs.setdefault("command",self.export_results)
        btn = tk.Button(self, cnf, **btn_kwargs)

        # create entry for filename and populate with default
        entry_kwargs = kwargs.get("entry_kwargs", {})
        #entry_kwargs.setdefault()
        entry = tk.Entry(self, cnf,**entry_kwargs)

        # pack away!
        btn.grid(row=0, column=0, sticky="NSWE")
        entry.grid(row=0, column=1, sticky="NSWE")

        # change column weights
        self.columnconfigure(1,weight=1)

        # store
        self.widgets.buttons["save"] = btn
        self.widgets.entries["save"] = entry

        # store command
        self.command = command

        # store file type to save to 
        self.ftype = ftype

        # store fname default
        if fname is None:
            fname = "output.txt"
        entry.insert(0,fname)

    def save(self):
        self.export_results()

    def export_results(self):
        command = self.command
        if callable(command):
            results = command()
        else:
            results = command
        ftype = self.ftype


        # get current entry filename
        fname = self.widgets.entries["save"].get()

        if ftype is None or ftype.upper() == "CSV":
            if fname is None:
                fname = "output.csv"

            with open(fname, "w") as outfile:
                header = []
                values = []
                for key in results.keys():
                    unit = results[key]["unit"]
                    string = key+" - ["+unit+"]"
                    header.append(string)
                for key in results.keys():
                    value = str(results[key]["value"])
                    values.append(value)

                writer = csv.writer(outfile)

                writer.writerow(header)
                writer.writerow(values)

        elif ftype.upper() == "JSON":
            if fname is None:
                fname = "output.json"

            with open(fname, "w") as outfile:
                json.dump(results, outfile) 
        elif ftype.upper() == "PNG":
            if fname is None:
                fname = "output.png"
            self.command()
        else:
            raise ValueError("Not supported export type. Try:\n") # add list of types
        
        # print
        print("saved to '%s'"%(fname))
    
    def get(self):
        return self.widgets.entries["save"].get()



class ResultOutput(tk.Frame):
    def __init__(self, key, value=None, unit=None, master=None, cnf=None, *args, latex=True, **kwargs):
        btn_kwargs = kwargs.pop("btn_kwargs", {})
        super().__init__(master, cnf, *args, **kwargs)

        # create widgets
        self.widgets = Widgets()

        # create button/label of property name to display
        if latex is True:
            key = self.convert_to_latex(key)
        btn_kwargs.setdefault("text", key)
        # define later to call new window with settings
        btn_kwargs.setdefault("command", None)
        btn = tk.Button(self, cnf, **btn_kwargs)

        # create label of value
        value_lbl_kwargs = kwargs.get("value_lbl_kwargs", {})
        value_lbl_kwargs.setdefault("text", value)
        value_lbl_kwargs.setdefault("bg", "white")
        value_lbl = tk.Label(self, cnf, **value_lbl_kwargs)

        # create unit label
        unit_lbl_kwargs = kwargs.get("unit_lbl_kwargs", {})
        unit_lbl_kwargs.setdefault("text", unit)
        unit_lbl = tk.Label(self, cnf, **unit_lbl_kwargs)

        # pack away!
        btn.grid(row=0, column=0, sticky="NSWE")
        value_lbl.grid(row=0, column=1, sticky="NSWE")
        unit_lbl.grid(row=0, column=2, sticky="NSWE")

        # save to self
        self.widgets.buttons["key"] = btn   # not sure why its a str
        self.widgets.labels["value"] = value_lbl
        self.widgets.labels["unit"] = unit_lbl

        # configure columns size
        self.columnconfigure(0, weight=2, uniform="fred")
        self.columnconfigure(1, weight=2, uniform="fred")
        self.columnconfigure(2, weight=1, uniform="fred")

    def update_text(self, text):
        self.widgets.labels["value"].config(text=text)

    def updata_value_label(self, key, value):
        kwargs = {key: value}
        self.widgets.labels["value"].config(**kwargs)

    def convert_to_latex(self, input_string: str):

        input_string = input_string.replace("lambda", "\u03BB")
        #input_string = input_string.replace("sheath", "r_p/\u03BB_De")

        split_string = input_string.split("_")
        if len(split_string) > 1:
            text = []
            for k, string in enumerate(split_string):
                if k == 0:
                    text.append(string)
                else:
                    text.append(string)
            text = "".join(text)
        else:
            text = split_string[0]

        return text



import tkinter as tk

from pstl.gui import Widgets
from pstl.gui.langmuir.results import SingleProbeLangmuirResultsFrame as SPLResultFrame
from pstl.gui.langmuir.control import SingleProbeLangmuirResultsControlFrame as SPLControlFrame

from pstl.diagnostics.probes.langmuir.single.helpers import available_plasma_properties


class SingleProbeLangmuirPanelFrame(tk.Frame):
    """
    This class creates one frame that contains the results page and the page
    that contains subframes of controls for each results button
    """
    def __init__(
            self,
            *args,
            master  :   tk.Frame | None =   None,
            cnf     :   dict            =   {},
            **kwargs):
        controls_frame_kwargs = kwargs.pop("controls_kwargs",{})
        results_frame_kwargs = kwargs.pop("displays_kwargs",{})
        super().__init__(master,cnf,*args,**kwargs)

        # add widgets organizer
        self.widgets = Widgets()

        # create controls
        controls_frame_kwargs.setdefault("master", self)
        controls_frame_kwargs.setdefault("cnf", cnf)
        controls_frame = SPLControlFrame(**controls_frame_kwargs)
        # pack it away!
        controls_frame.grid(
            row=0,
            column=0,
            columnspan=2,
            sticky="NWSE",
        )

        # create results
        results_frame_kwargs.setdefault("master", self)
        results_frame_kwargs.setdefault("cnf", cnf)
        results_frame = SPLResultFrame(**results_frame_kwargs)
        # pack it away!
        results_frame.grid(
            row=0,
            column=0,
            columnspan=2,
            sticky="NWSE",
        )

        # create display and contrl frame buttons
        display_btn = tk.Button(self,cnf,text="Displays",command=self.raise_display_frame)
        display_btn.grid(row=1,column=0,sticky="WSE")
        control_btn = tk.Button(self,cnf,text="Controls",command=self.raise_control_frame)
        control_btn.grid(row=1,column=1,sticky="WSE")

        # save to widgets
        self.widgets.frames["control"] = controls_frame
        self.widgets.frames["display"] = results_frame
        self.widgets.buttons["control"] = control_btn
        self.widgets.buttons["display"] = display_btn

        # save handle access
        self._controls = controls_frame
        self._displays = results_frame

        # configure buttons to ew
        self.set_display_btn_default()

    @property
    def controls(self):
        return self._controls
    
    @property
    def displays(self):
        return self._displays
    
    def save(self):
        self.displays.save()

    def get_updated_algorithm_kwargs_from_control_panel(self):
        """
        Returns dict of i.e. {"V_f_kwargs: value, ...}
        """
        algorithm_kwargs = self.controls.get_all_controls_entries()
        return algorithm_kwargs

    def set_display_btn_default(self):
        for key in available_plasma_properties:
            raise_func=self.get_raise_func(key)
            #self.displays.widgets.frames["results"][key].widgets.buttons["key"].configure(command=self.controls.widgets.frames["controls"][key].tkraise)
            #self.displays.widgets.frames["results"][key].widgets.buttons["key"].configure(command=self.raise_control_frame)
            self.displays.widgets.frames["results"][key].widgets.buttons["key"].configure(command=raise_func)


    def get_raise_func(self, key):
        if key == "V_f":
            raise_func = self.raise_func_floating_potential
        elif key == "V_s":
            raise_func = self.raise_func_plasma_potential
        elif key == "I_is":
            raise_func = self.raise_func_ion_saturation_current
        elif key == "n_i":
            raise_func = self.raise_func_ion_density
        elif key == "I_es":
            raise_func = self.raise_func_electron_saturation_current
        elif key == "KT_e":
            raise_func = self.raise_func_electron_temperature
        elif key == "n_e":
            raise_func = self.raise_func_electron_density
        elif key == "J_es":
            raise_func = self.raise_func_electron_saturation_current_density
        elif key == "J_is":
            raise_func = self.raise_func_ion_saturation_current_density
        elif key == "lambda_De":
            raise_func = self.raise_func_lambda_De
        elif key == "r_p/lambda_De":
            raise_func = self.raise_func_ratio
        elif key == "sheath":
            raise_func = self.raise_func_sheath
        else:
            raise_func = self.raise_control_frame

        return raise_func



    def raise_control_frame_keyed(self, key):
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()

    def raise_control_frame(self):
        self.controls.tkraise()
    def raise_display_frame(self):
        self.displays.tkraise()

    def raise_func_floating_potential(self):
        key="V_f"
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()
    def raise_func_plasma_potential(self):
        key="V_s"
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()
    def raise_func_ion_saturation_current(self):
        key="I_is"
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()
    def raise_func_ion_saturation_current_density(self):
        key="J_is"
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()
    def raise_func_electron_saturation_current(self):
        key="I_es"
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()
    def raise_func_electron_saturation_current_density(self):
        key="J_es"
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()
    def raise_func_electron_density(self):
        key="n_e"
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()
    def raise_func_ion_density(self):
        key="n_i"
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()
    def raise_func_electron_temperature(self):
        key="KT_e"
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()
    def raise_func_lambda_De(self):
        key="lambda_De"
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()
    def raise_func_ratio(self):
        key="r_p/lambda_De"
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()
    def raise_func_sheath(self):
        key="sheath"
        self.raise_control_frame()
        self.controls.widgets.frames["controls"][key].tkraise()


from abc import ABC, abstractmethod

import numpy as np


from pstl.utls.objects import setup as object_setup
from pstl.diagnostics.probes.classes import Probe

available_probe_classes = [
    [4, "4grid", "four", "fourgrid"],
    [ 3, "3grid", "three","threegrid"],
]

def setup(settings, *args, **kwargs):
    """
    Creates and returns a RPA probe object  based on settings dictionary passed in.
    The settings parameter must have keys 'ngrids'.
    
    Keys:
        'ngrids'     : int   ->  number of grids [4,3]
    (optional)
        'name'      : str   ->  name designation for probe 
        'args'      : tuple ->  addional position arguments
        'kwargs'    : dict  ->  addional keyword arguments

    Returns: Probe Object
    """
    # determine shape of 
    shape = settings.pop("ngrids").lower()
    if shape in available_probe_classes[0]:
        Probe = FourGridRPA
    elif shape in available_probe_classes[1]:
        Probe = ThreeGridRPA
    else:
        raise ValueError("'%s' is not a valid option."%(shape))
    
    output_object: RepellingPotentialAnalyzer = object_setup(
        *args,
        settings=settings,
        builders={},
        Builder=Probe,
    #    to_args=
        **kwargs,
    )

    return output_object

class RepellingPotentialAnalyzer(Probe, ABC):
    def __init__(self, *args, ngrids: int | None = None, name=None, description=None, **kwargs) -> None:
        # *args should be the potentials of the grids
        # need to add something for space between grids & collection area 
        # to determine resounce times
        self._ngrids = int(ngrids) if ngrids is not None else int(len(args))

    @property
    def ngrids(self):
        return self._ngrids
    

class FourGridRPA(RepellingPotentialAnalyzer):
    def __init__(self, *args, name=None, description=None, **kwargs) -> None:
        ngrids = 4
        super().__init__(*args, ngrids=ngrids, name=name, description=description, **kwargs)

class ThreeGridRPA(RepellingPotentialAnalyzer):
    def __init__(self, *args, name=None, description=None, **kwargs) -> None:
        ngrids = 3
        super().__init__(*args, ngrids=ngrids, name=name, description=description, **kwargs)

class RPA(RepellingPotentialAnalyzer):
    def __init__(self, *args, ngrids: int | None = None, name=None, description=None, **kwargs) -> None:
        super().__init__(*args, ngrids=ngrids, name=name, description=description, **kwargs)



import os 
import tkinter as tk

from PIL import ImageTk, Image

path = os.path.dirname(__file__)

pstl_ico_path = os.path.join(path, "pstl.ico")
pstl_png_path = os.path.join(path, "pstl.png")
pstl_16_ico_path = os.path.join(path, "pstl-16.ico")
pstl_32_ico_path = os.path.join(path, "pstl-32.ico")



import numpy as np
import numpy.typing as npt

from pstl.utls.helpers import ideal_gas_law_pressure_to_density

class Xenon():
    def __init__(
            self, 
            E:int | float, n: int | float, z: int | float, js:npt.ArrayLike, *args, 
            Tgas: int | float = 300, pressure: bool = False,
            **kwargs):
        """
        This class solves for correction factors and the curent densities at thruster exit plane using
        E: energy of ions in eV
        n: background neutral density in m^-3 (current version assumes uniform constant density) if pressure is True, then it is assumed n to be given as pressure in Torr
        z: is disntance downstream of thruster exit plane in m
        js: is an arraylike of current densities of single, double, [triple] given in arbitry units
        Attributes:
        dictionaries start with keys of 1 till 3 for single, doubles, triples respectivley.
        js: dict of current densities provided at distance z downstream
        j0s: dict of corrected current densities at thruster exit plane
        cfs: dict of correction factors 

        Reference source:
        Rohit Shastry, Richard R. Hofer, Bryan M. Reid, Alec D. Gallimore; Method for analyzing ExB probe spectra from Hall thruster plumes. Rev. Sci. Instrum. 1 June 2009; 80 (6): 063502. https://doi.org/10.1063/1.3152218
        """
        self.z = z
        if pressure is True:
            n = ideal_gas_law_pressure_to_density(n,T_gas=Tgas)
            self.n = n
        elif pressure is False:
            self.n = n
        self.E = E

        self.js = self._organize_js(js)
        self.sigmas = self.calc_sigmas(E)
        self.cfs = self.calc_correction_factors(n, z)
        self.j0s = self.correct_current_densitites(self.js)


    def _organize_js(self,js):
        js_new = {}
        if isinstance(js, dict):
            for species, j in js.items():
                js_new[species] = j
        else:
            for k, j in enumerate(js):
                species = k+1
                js_new[species] = j
        return js_new

    def calc_sigmas(self,E):
        sigmas = {}
        sigmas[1] = self.calc_sigma1(E)
        sigmas[2] = self.calc_sigma2(E)
        sigmas[3] = self.calc_sigma3()
        sigmas[4] = self.calc_sigma4(E)
        sigmas[5] = self.calc_sigma5(E)
        return sigmas

    def calc_sigma1(self, E):
        # Where E is beam ion energy in eV
        return (87.3-13.6*np.log10(E))*1e-20
    def calc_sigma2(self, E):
        # Where E is beam ion energy in eV
        return (45.7-8.9*np.log10(E))*1e-20
    def calc_sigma3(self):
        return 2*1e-20
    def calc_sigma4(self, E):
        # Where E is beam ion energy in eV
        return self.calc_sigma1(E)
    def calc_sigma5(self, E):
        # Where E is beam ion energy in eV
        return (16.9-3*np.log10(E))*1e-20
        
    def calc_j1_j10(self,n,z):
        # n: density
        # E: Beam ion energy in eV
        # z: distance in meters from thruster plane to probe
        nz = np.multiply(n,z)
        term1 = np.multiply(-nz, self.sigmas[1])
        j1_j10 = np.exp(term1)
        return j1_j10
    def calc_j2_j20(self,n,z): # make this a regular func that does not need n,z
        # n: density
        # E: Beam ion energy in eV
        # z: distance in meters from thruster plane to probe
        nz = np.multiply(n,z)
        term1 = np.multiply(-nz, np.add(self.sigmas[2],self.sigmas[3]))
        j2_j20 = np.exp(term1)
        return j2_j20
    def calc_j3_j20(self,n,z): # make this a regular func that does not need n,z
        # n: density
        # E: Beam ion energy in eV
        # z: distance in meters from thruster plane to probe
        nz = np.multiply(n,z)
        term1 = np.exp(np.multiply(-nz,self.sigmas[4]))
        term2 = np.exp(np.multiply(-nz,np.add(self.sigmas[2],self.sigmas[3])))
        numerator = np.subtract(term1,term2)
        denominator = np.sum([self.sigmas[2],self.sigmas[3],-self.sigmas[4]])
        j3_j20 = np.multiply(
            self.sigmas[3]/2,
            np.divide(numerator,denominator)
        )
        return j3_j20
    
    def calc_correction_factors(self, n, z):
        cf = {}
        cf[1] = self.calc_correction_factor_for_singles(n, z)
        cf[2] = self.calc_correction_factor_for_doubles(n, z)
        cf[3] = self.calc_correction_factor_for_triples(n, z)
        return cf
    
    def calc_correction_factor_for_singles(self, n, z):
        cf = self.calc_j1_j10(n,z)
        return cf
    def calc_correction_factor_for_doubles(self, n, z):
        term1 = self.calc_j2_j20(n,z)
        term2 = self.calc_j3_j20(n,z)
        cf = np.add(term1, term2)
        return cf
    def calc_correction_factor_for_triples(self, n, z):
        nz = np.multiply(n,z)
        cf = np.exp(np.multiply(-nz,self.sigmas[5]))
        return cf
    
    def correct_current_densitites(self, js):
        j0s = {}
        if isinstance(js, dict):
            for species, j in js.items():
                j0s[species] = np.divide(j,self.cfs[species])
        else:
            for k, j in enumerate(js):
                species = k+1
                j0s[species] = np.divide(j,self.cfs[species])
        return j0s



from pstl.utls.helpers import method_function_combiner
from pstl.utls.helpers import method_selector 

from .shape_methods import cylinderical_method, spherical_method, planar_method


def shape_selector(shape):
    # select function based on integer_method
    func = method_selector(shape, available_methods, available_functions)
    return func

# Declare available methods
available_methods = {
    0: ['cylinderical', 'cylindrical'],
    1: ['spherical'],
    2: ['planar', 'planer'],
}
# Declare correspondin functions for available_methods
available_functions = {
    0: cylinderical_method,
    1: spherical_method,
    2: planar_method,
}

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
options = method_function_combiner(available_methods, available_functions)


import tkinter as tk

import pandas as pd
import numpy as np
from scipy.signal import savgol_filter
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.backend_bases import key_press_handler
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk  # type: ignore

from pstl.gui import WidgetOrganizer, Widgets
from pstl.gui.langmuir.results import SaveResults

class MatplotlibCanvas(tk.Frame):
    def __init__(self, master=None, cnf=None, *args, fig: Figure | None = None, ax: Axes | None = None,  **kwargs):
        width = kwargs.pop("width", 8)
        height = kwargs.pop("height", 6)
        dpi = kwargs.pop("dpi", 100)
        super().__init__(master, cnf, *args, **kwargs)

        # add widgets
        self.widgets = Widgets()

        if fig is None:
            # initialize matplotlib figure and axes

            self.fig = Figure(figsize=(width, height), dpi=dpi)
            self.fig.set_tight_layout(True)
        else:
            self.fig = fig

        if ax is None:
            self.ax = self.fig.add_subplot(
                1, 1, 1)  # type: ignore
        else:
            self.ax = ax

        # A tk.DrawingArea.
        self.canvas = FigureCanvasTkAgg(self.fig, master=self)
        self.canvas.draw()

        # Pack Canvas
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)

    def set_fig_to_linear(self):
        fig = self.fig

        ax = self.ax
        scale = ax.get_yscale()
        if len(ax.lines) == 0:
            ax.set_yscale("linear")
        else:
            ydata = ax.lines[0].get_ydata()
            ydata_minimum = np.min(ydata)
            ydata_maximum = np.max(ydata)
            ylim = ax.get_ylim()

            ypad = 0.05*(np.max(ydata)-np.min(ydata))

            ylim = [ydata_minimum-ypad, ydata_maximum+ypad]

            ax.set_yscale("linear")

            ax.set_ylim(ylim[0], ylim[1])
        fig.canvas.draw()

    def set_fig_to_semilogy(self):
        fig = self.fig

        ax = self.ax
        scale = ax.get_yscale()

        if len(ax.lines) == 0:
            ax.set_yscale("log")
        else:
            ydata = ax.lines[0].get_ydata()

            # Set the y-axis limits with minimum positive value padding
            y_lim = ax.get_ylim()
            y_min = np.floor(np.log10(np.min(ydata[ydata > 0])))
            y_max = np.ceil(np.log10(y_lim[1]))

            ax.set_yscale("log")

            ax.set_ylim(10**(y_min), 10**y_max)
        fig.canvas.draw()


class MatplotlibCanvasWToolbar(MatplotlibCanvas):
    def __init__(self, master=None, cnf=None, *args, **kwargs):
        super().__init__(master, cnf, *args, **kwargs)
        # pack_toolbar=False will make it easier to use a layout manager later on.
        self.toolbar = NavigationToolbar2Tk(
            self.canvas, self, pack_toolbar=False)
        self.toolbar.update()

        # Pack Toolbar
        self.toolbar.pack(side=tk.TOP, fill=tk.X, expand=True)


class MatplotlibCanvasWToolbarSave(MatplotlibCanvasWToolbar):
    def __init__(self, master=None,
                 cnf={}, *args,
                 fig: Figure | None = None, ax: Axes | None = None,
                 saveas: str = "figure1.png", stype="png",
                 **kwargs):
        super().__init__(master, cnf, *args, fig=fig, ax=ax, **kwargs)

        self.saveas = saveas
        self.args = args
        self.kwargs = kwargs

        # add save default button
        btn = SaveResults(self.save,fname=saveas,ftype=stype,master=self,cnf=cnf)
        self.widgets.frames['save'] = btn
        #    self, text="Save", command=self.save)
        #self.widgets.buttons['save'].pack(
        btn.pack(side=tk.TOP, fill=tk.X, expand=True)
        

    def save(self, **kwargs):#, saveas: str | None = None, **kwargs):
        #if saveas is None:
        #    saveas = self.widgets.frames['save'].get()
        saveas = self.widgets.frames["save"].get()
        self.fig.savefig(saveas, **kwargs)
        print("save to '%s'"%(saveas))


class LinearSemilogyCanvas(MatplotlibCanvasWToolbarSave):
    def __init__(self, master=None,
                 cnf={}, fig: Figure | None = None, ax: Axes | None = None,
                 saveas: str = "figure1.png",
                 **kwargs):
        super().__init__(master, cnf, fig=fig, ax=ax, saveas=saveas, **kwargs)

        self.widgets.buttons['linear'] = tk.Button(
            self, text="Linear", command=self.set_fig_to_linear
        )
        self.widgets.buttons['semilogy'] = tk.Button(
            self, text="Semilogy", command=self.set_fig_to_semilogy
        )

        # pack linear and semilogy buttons
        self.widgets.buttons['linear'].pack(
            side=tk.LEFT, fill=tk.BOTH, expand=True
        )
        self.widgets.buttons['semilogy'].pack(
            side=tk.RIGHT, fill=tk.BOTH, expand=True
        )

class FirstDerivativeSingleProbeLangmuir(MatplotlibCanvasWToolbarSave):
    def __init__(
            self, 
            master=None, 
            cnf={}, 
            *args, 
            fig: Figure | None = None, 
            ax: Axes | None = None, 
            saveas: str = "first_derivative.png", 
            stype="png", 
            **kwargs):
        kwargs.setdefault("width", 5)
        kwargs.setdefault("height", 4)
        kwargs.setdefault("dpi", 100)
        
        super().__init__(master, cnf, *args, fig=fig, ax=ax, saveas=saveas, stype=stype, **kwargs)

        self._state = False

    def solve(self,solver):
        # Try to solve and plot Frist Derivative
        try:
            # pass in the data
            voltage = solver.data.voltage
            current = solver.data.current

            # try to get current_e
            try:
                voltage_e = solver.data.voltage
                current_e = solver.data.current_e
                data={
                "voltage": voltage, 
                "current": current, 
                "current_e": current_e,
                "voltage_e": voltage_e,
                }
            except:
                data={
                "voltage": voltage, 
                "current": current, 
                }

            # change to pandas for sorting and removal of data for derivative
            data = pd.DataFrame(data)
            # sort data
            data.sort_values(by="voltage", ascending=True,inplace=True)
            # remove duplicates
            data.drop_duplicates("voltage", keep="last", inplace=True)
            # unpack back out
            voltage = data.voltage.to_numpy()
            current = data.current.to_numpy()
            try:
                current_e = data.current_e.to_numpy()
                voltage_e = data.voltage_e.to_numpy()
            except:
                pass

            try:
                # solve for floating potential
                vf = solver.results["V_f"]["value"]
                if vf is None:
                    vf = voltage[np.where(current>=0)[0][0]]
            except:
                vf = voltage[np.where(current>=0)[0][0]]
            finally:
                try:
                    indices = np.where(voltage>=vf)[0][0]
                except Exception as e:
                    raise ValueError(f"Value for V_f in results failed.\n{e}")

            # basic filter settings
            win_len = 6
            polyorder = 2

            # function for derivative
            def solve_1stdev(x, y):
                ## solve ##
                # dI/dV
                gradient_1D = np.gradient(y, x)
                # SG-dI/dV
                smoothed_1D = savgol_filter(
                    y,
                    window_length=win_len,
                    polyorder=polyorder)
                smoothed_1D = np.gradient(smoothed_1D, x)
                smoothed_1D_smoothed = savgol_filter(
                    smoothed_1D,
                    window_length=win_len,
                    polyorder=polyorder)
                # SGF-dI/dV
                smoothed_1D_savgol = savgol_filter(
                    y, 
                    window_length=win_len,
                    polyorder=polyorder,
                    deriv=1
                )
                # normalized by max
                gradient_1D = np.divide(gradient_1D, np.max(gradient_1D))
                smoothed_1D = np.divide(smoothed_1D, np.max(smoothed_1D))
                smoothed_1D_smoothed = np.divide(smoothed_1D_smoothed, np.max(smoothed_1D_smoothed))
                smoothed_1D_savgol = np.divide(smoothed_1D_savgol, np.max(smoothed_1D_savgol))
                return gradient_1D, smoothed_1D, smoothed_1D_smoothed, smoothed_1D_savgol
            ## solve 1st der ##
            # solve Ip der #
            (
                c_gradient_1D, 
                c_smoothed_1D, 
                c_smoothed_1D_smoothed, 
                c_smoothed_1D_savgol,
                ) = solve_1stdev(voltage, current)
            
            ## plot ##
            # set x and y labels
            self.ax.set_xlabel("Voltage [V]")
            self.ax.set_ylabel(r"$(dI_p/dV,dI_e/dV)/(dI_p/dV,dI_e/dV)_{max}$ [-]")
            
            # plot first der Ip
            self.ax.plot(voltage, c_gradient_1D, "--",color="C0",label=r"$\frac{dI_p}{dV}$", alpha=0.4)
            self.ax.plot(voltage, c_smoothed_1D,"--",color="C2",label=r"SG-$\frac{dI_p}{dV}$", alpha=0.4)
            self.ax.plot(voltage, c_smoothed_1D_smoothed,"--",color="C1",label=r"SG-$\frac{dI_p}{dV}$-SG")
            self.ax.plot(voltage, c_smoothed_1D_savgol,"--",color="C3",label=r"$\frac{dI_p}{dV}$ via SGF", alpha=0.7)
            
            # legend order if current_e not present
            order = [0,1,2,3]
            
            # try to add current_e stuff
            try:
                # get updated current_e and voltage_e based on Vf
                current_e = current_e[indices:]
                voltage_e = voltage_e[indices:]
                # solve Ie der #
                (
                    ce_gradient_1D, 
                    ce_smoothed_1D, 
                    ce_smoothed_1D_smoothed, 
                    ce_smoothed_1D_savgol,
                    ) = solve_1stdev(voltage_e, current_e)
                # plot first der Ie
                self.ax.plot(voltage_e, ce_gradient_1D, "-",color="C0",label=r"$\frac{dI_e}{dV}$", alpha=0.4)
                self.ax.plot(voltage_e, ce_smoothed_1D,"-",color="C2",label=r"SG-$\frac{dI_e}{dV}$", alpha=0.4)
                self.ax.plot(voltage_e, ce_smoothed_1D_smoothed,"-",color="C1",label=r"SG-$\frac{dI_e}{dV}$-SG")
                self.ax.plot(voltage_e, ce_smoothed_1D_savgol,"-",color="C3",label=r"$\frac{dI_e}{dV}$ via SGF", alpha=0.7)
                
                # legend order if current_e
                order = [6, 2]
            except:
                pass

            # add legend
            lns = self.ax.lines
            labs = [l.get_label() for l in lns]
            lgnd = self.ax.legend([lns[i] for i in order], [labs[i] for i in order])
            # make tight layout
            self.fig.tight_layout()
    
        except Exception as e:
            print("Unable to make first derivative plot:")
            print(e)
            pass
        except:
            print("unable to make first derivative plot")
            pass
    
    @property
    def state(self):
        return self._state
    @state.setter
    def state(self, change_to):
        if isinstance(change_to, bool):
            self._state = change_to
        else:
            raise TypeError("Must be a bool")

class LinearSemilogyDoubleCanvas(tk.Frame):
    def __init__(self, master=None, cnf=None, saveas="figures.png", sappends=["_linear", "_semilogy"], n=2, sharex=True, **kwargs):
        width = kwargs.pop("width", 5)
        height = kwargs.pop("height", 4)
        dpi = kwargs.pop("dpi", 100)
        stype = kwargs.pop("stype", "png")
        figure_kw = {'width': width, 'height': height, 'dpi': dpi, 'stype':stype}

        super().__init__(master, cnf, **kwargs)

        if isinstance(saveas, list):
            self.saveas = saveas
        elif isinstance(saveas, str):
            slist = saveas.split('.')
            if len(slist) == 1:
                path = saveas
                ext = str()
            elif len(slist) >= 2:
                num_ext_char = len(slist[-1])
                iend = -(num_ext_char+1)
                ext = saveas[iend:]
                path = saveas[:iend]
            elif len(slist) == 0:
                raise ValueError("Something went wrong")
            else:
                raise ValueError("Lenth of list cannot be negative")

            saveas = []
            for i in range(n):
                if i == 0:
                    sappend = "_linear"
                elif i == 1:
                    sappend = "_semilogy"
                else:
                    sappend = str(i)
                saveas.append(path+sappend+ext)

        self.widgets = Widgets()

        self.widgets.frames['linear'] = MatplotlibCanvasWToolbarSave(
            self, saveas=saveas[0], **figure_kw)

        self.widgets.frames['semilogy'] = MatplotlibCanvasWToolbarSave(
            self, saveas=saveas[1], **figure_kw)

        # remap ax for plotting
        self.frame1 = self.widgets.frames['linear']
        self.ax1 = self.frame1.ax
        self.frame2 = self.widgets.frames['semilogy']
        self.ax2 = self.frame2.ax
        # set frame2 to semilogy
        self.widgets.frames['semilogy'].set_fig_to_semilogy()

        if sharex:
            self.ax2.sharex(self.ax1)

        # pack
        self.frame1.grid(row=0, column=0, sticky="NSWE")
        self.frame2.grid(row=0, column=1, sticky="NSWE")

        # add save default button
        self.widgets.buttons['save_both'] = tk.Button(
            self, text="Save Both", command=self.save_both)
        self.widgets.buttons['save_both'].grid(
            row=1, column=0, columnspan=2, sticky="NSEW")

    def save_both(self, **kwargs):

        self.frame1.save()
        self.frame2.save()

    def save(self):
        self.save_both()

    @staticmethod
    def update_both(func, *args, **kwargs):
        def decorator(self, *args, **kwargs):
            func(*args, **kwargs)
            self.frame1.set_fig_to_linear()
            self.frame2.set_fig_to_semilogy()
        return decorator  # (*args, **kwargs)

    @update_both
    @staticmethod
    def add_plot(ax_func, *args, **kwargs):
        ax_func(*args, **kwargs)

    @update_both
    @staticmethod
    def add_to_both(ax1_func, ax2_func, *args, **kwargs):
        ax1_func(*args, **kwargs)
        ax2_func(*args, **kwargs)

    def legend(self, ax1_bool=True, ax2_bool=True, *args, **kwargs):
        if ax1_bool:
            self.ax1.legend(*args, **kwargs)
        if ax2_bool:
            self.ax2.legend(*args, **kwargs)

    def add_raw_data(self, voltage, current, *args, **kwargs):
        # set defaults
        kwargs.setdefault("label", "Raw Data")
        kwargs.setdefault("color", "C0")
        kwargs.setdefault("marker", "s")
        kwargs.setdefault("markerfacecolor", "none")
        kwargs.setdefault("alpha", 0.5)

        # save data
        self.raw_data = pd.DataFrame({'voltage': voltage, 'current': current})

        # add plot
        self.add_to_both(self.ax1.plot, self.ax2.plot,
                         voltage, current, *args, **kwargs)

    def add_data(self, voltage, current, *args, **kwargs):
        # set defaults
        kwargs.setdefault("label", "Experimental Data")
        kwargs.setdefault("color", "C0")
        kwargs.setdefault("marker", "^")
        kwargs.setdefault("markerfacecolor", "none")

        # save data
        self.data = pd.DataFrame({'voltage': voltage, 'current': current})

        # add plot
        self.add_to_both(self.ax1.plot, self.ax2.plot,
                         voltage, current, *args, **kwargs)

    def add_deleted_points(self, voltage, current, *args, **kwargs):
        kwargs.setdefault("label", "Removed Data")
        kwargs.setdefault("color", "C1")
        kwargs.setdefault("linestyle", "none")
        kwargs.setdefault("marker", "x")
        self.deleted_points = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_to_both(self.ax1.plot, self.ax2.plot,
                         voltage, current, *args, **kwargs)

    def add_filtered_data(self, voltage, current, *args, **kwargs):
        kwargs.setdefault("label", "Filtered Data")
        kwargs.setdefault("color", "C7")
        kwargs.setdefault("marker", "v")
        kwargs.setdefault("markerfacecolor", "none")
        self.filtered_data = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_to_both(self.ax1.plot, self.ax2.plot,
                         voltage, current, *args, **kwargs)

    def add_smoothed_data(self, voltage, current, *args, **kwargs):
        kwargs.setdefault("label", "Smoothed Data")
        kwargs.setdefault("color", "C8")
        kwargs.setdefault("marker", "o")
        kwargs.setdefault("markerfacecolor", "none")
        self.smoothed_data = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_to_both(self.ax1.plot, self.ax2.plot,
                         voltage, current, *args, **kwargs)


class LinearSemilogyDoubleCanvasSingleProbeLangmuir(LinearSemilogyDoubleCanvas):
    def __init__(self, master=None, cnf=None, saveas="figures.png", sappends=["_linear", "_semilogy"], n=2, sharex=True, **kwargs):
        super().__init__(master, cnf, saveas, sappends, n, sharex, **kwargs)

        self.raw_data = None
        self.filtered_data = None
        self.deleted_points = None
        self.smoothed_data = None

        self.vf = None
        self.vs = None

        self.electron_retarding_fits = []
        self.electron_retarding_fills = []

        self.electron_saturation_fit = None
        self.electron_saturation_fill = None

        self.ion_saturation_fit = None
        self.ion_saturation_fill = None

        # make 1der plot show hide buttons
        if isinstance(saveas, str):
            slist = saveas.split('.')
            if len(slist) == 1:
                path = saveas
                ext = str()
            elif len(slist) >= 2:
                num_ext_char = len(slist[-1])
                iend = -(num_ext_char+1)
                ext = saveas[iend:]
                path = saveas[:iend]
            elif len(slist) == 0:
                raise ValueError("Something went wrong")
            else:
                raise ValueError("Lenth of list cannot be negative")

            saveas_der = []
            for i in range(1):
                if i == 0:
                    sappend = "_first_derivative"
                else:
                    sappend = str(i)
                saveas_der.append(path+sappend+ext)

        canvas_1der = FirstDerivativeSingleProbeLangmuir(self,cnf, saveas=saveas_der, **kwargs)
        canvas_1der.grid(row=0, column=0, sticky="NWSE")
        self.widgets.frames["1der"] = canvas_1der
        show_hide_btn = tk.Button(self,cnf,text="Show/Hide",command=self.show_hide)
        show_hide_btn.grid(row=2, column=0,columnspan=2,sticky="NSEW")
        self.widgets.frames["linear"].tkraise()
        self.frame3 = self.widgets.frames["1der"]
        self.ax3 = canvas_1der.ax
    def show_hide(self):
        canvas_1der = self.widgets.frames["1der"]
        canvas_linear = self.widgets.frames["linear"]
        state = canvas_1der.state
        if state is True: # is showing thus hide
            canvas_linear.tkraise()
            state = False
        elif state is False: # not showing
            canvas_1der.tkraise()
            state = True

        canvas_1der.state = state


    def set_xlabel_linear(self, label=None, *args, **kwargs):
        label = r"$V_{bias}\ [V]$" if label is None else label
        self.ax1.set_xlabel(label)

    def set_xlabel_semilogy(self, label=None, *args, **kwargs):
        label = r"$V_{bias}\ [V]$" if label is None else label
        self.ax2.set_xlabel(label)

    def set_xlabel(self, label_linear=None, label_semilogy=None, *args, **kwargs):
        self.set_xlabel_linear(label=label_linear)
        self.set_xlabel_semilogy(label=label_semilogy)

    def set_ylabel_linear(self, label=None, *args, **kwargs):
        label = r"$I_{probe}\ [A]$" if label is None else label
        self.ax1.set_ylabel(label)

    def set_ylabel_semilogy(self, label=None, *args, **kwargs):
        label = r"$I_{e}\ [A]$" if label is None else label
        self.ax2.set_ylabel(label)

    def set_ylabel(self, label_linear=None, label_semilogy=None, *args, **kwargs):
        self.set_ylabel_linear(label=label_linear)
        self.set_ylabel_semilogy(label=label_semilogy)

    def add_probe_trace(self, voltage, current, *args, **kwargs):
        kwargs.setdefault("label", "Experimental Data")
        kwargs.setdefault("color", "C0")
        kwargs.setdefault("marker", "^")
        kwargs.setdefault("markerfacecolor", "none")
        # save data
        self.probe_trace = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_plot(self.ax1.plot,
                      voltage, current, *args, **kwargs)
        self.set_xlabel_linear()
        self.set_ylabel_linear()

    def add_electron_trace(self, voltage, current, *args, **kwargs):
        kwargs.setdefault("label", "Experimental Data")
        kwargs.setdefault("color", "C0")
        kwargs.setdefault("marker", "^")
        kwargs.setdefault("markerfacecolor", "none")
        # save data
        self.electron_trace = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_plot(self.ax2.plot,
                      voltage, current, *args, **kwargs)
        self.set_xlabel_semilogy()
        self.set_ylabel_semilogy()

    def add_semilogy_trace(self, voltage, current, *args, label=None,**kwargs):
        kwargs.setdefault("label", "Experimental Data")
        kwargs.setdefault("color", "C0")
        kwargs.setdefault("marker", "^")
        kwargs.setdefault("markerfacecolor", "none")
        # save data
        #self.electron_trace = pd.DataFrame({'voltage': voltage, 'current': current})
        self.add_plot(self.ax2.plot,
                      voltage, current, *args, **kwargs)
        label = r"$I_{probe}\ [A]$" if label is None else label
        self.set_xlabel_semilogy()
        self.set_ylabel_semilogy(label=label)

    def add_floating_potential(self, V_f, *args, **kwargs):
        kwargs.setdefault("label", rf"$V_{{f}} = {V_f:.2f}V$")
        kwargs.setdefault("color", "C2")
        self.V_f = V_f
        self.add_to_both(self.ax1.axvline, self.ax2.axvline,
                         V_f, *args, **kwargs)

    def add_plasma_potential(self, V_s, *args, **kwargs):
        kwargs.setdefault("label", rf"$V_{{s}} = {V_s:.2f}V$")
        kwargs.setdefault("color", "C4")
        self.V_s = V_s
        self.add_to_both(self.ax1.axvline, self.ax2.axvline,
                         V_s, *args, **kwargs)

    def add_electron_retarding_fit(self, voltage, current, KT_e, *args, **kwargs):
        kwargs.setdefault("label", rf"$KT_{{e}} = {KT_e:.2f}eV$")
        kwargs.setdefault("color", "C3")
        kwargs.setdefault("linestyle", "--")
        self.electron_retarding_fits.append(
            pd.DataFrame({'voltage': voltage, 'current': current}))
        self.add_plot(self.ax2.plot,
                      voltage, current, *args, **kwargs)

    def add_ion_saturation_fit(self, voltage, current, I_is, *args, **kwargs):
        kwargs.setdefault("label", rf"$I_{{is}} = {I_is:.2e}A$")
        kwargs.setdefault("color", "C6")
        kwargs.setdefault("linestyle", "--")
        self.ion_saturation_fit = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_plot(self.ax1.plot,
                      voltage, current, *args, **kwargs)

    def add_electron_saturation_fit(self, voltage, current, I_es, *args, **kwargs):
        kwargs.setdefault("label", rf"$I_{{es}} = {I_es:.2e}A$")
        kwargs.setdefault("color", "C5")
        kwargs.setdefault("linestyle", "--")
        self.electron_saturation_fit = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_plot(self.ax2.plot,
                      voltage, current, *args, **kwargs)

    def add_electron_retarding_fill(self, xstart, xstop, *args, **kwargs):
        kwargs.setdefault("color", "C3")
        kwargs.setdefault("alpha", 0.3)
        self.electron_retarding_fills.append([xstart, xstop])
        self.add_to_both(self.ax1.axvspan, self.ax2.axvspan,
                         xstart, xstop, *args, **kwargs)

    def add_electron_saturation_fill(self, xstart, xstop, *args, **kwargs):
        kwargs.setdefault("color", "C5")
        kwargs.setdefault("alpha", 0.3)
        self.electron_saturation_fill = [xstart, xstop]
        self.add_to_both(self.ax1.axvspan, self.ax2.axvspan,
                         xstart, xstop, *args, **kwargs)

    def add_ion_saturation_fill(self, xstart, xstop, *args, **kwargs):
        kwargs.setdefault("color", "C6")
        kwargs.setdefault("alpha", 0.3)
        self.ion_saturation_fill = [xstart, xstop]
        self.add_to_both(self.ax1.axvspan, self.ax2.axvspan,
                         xstart, xstop, *args, **kwargs)
    def add_1derivative(self, solver):
        self.widgets.frames["1der"].solve(solver)

    def make_plot(self, solver):
        app = self
        # make xdomain for fit plotting
        xdomain = [np.min(solver.data.voltage), np.max(solver.data.voltage)]

        # solve for KT_e fit data
        KT_e_fit = solver.results["KT_e"]["other"]["fit"]
        voltage_KT_e_fit = KT_e_fit.xrange(domain=xdomain)
        current_KT_e_fit = KT_e_fit(voltage_KT_e_fit)
        # solve for I_es fit data
        I_es_fit = solver.results["I_es"]["other"]["fit"]
        voltage_I_es_fit = I_es_fit.xrange(domain=xdomain)
        current_I_es_fit = I_es_fit(voltage_I_es_fit)

        # solve for I_is fit data
        I_is_fit = solver.results["I_is"]["other"]["fit"]
        voltage_I_is_fit = I_is_fit.xrange(domain=xdomain)
        current_I_is_fit = I_is_fit(voltage_I_is_fit)

        # Make Plots ###
        # add probe trace
        app.add_probe_trace(solver.data.voltage, solver.data.current)
        # add electron current
        app.add_electron_trace(solver.data.voltage, solver.data.current_e)
        # mark deleted points
        app.add_deleted_points(solver.deleted_data.voltage,
                               solver.deleted_data.current)

        # Add to Plasma Properties to Plots
        app.add_floating_potential(solver.results["V_f"]["value"])
        app.add_plasma_potential(solver.results["V_s"]["value"])

        # Electron Retarding fit and fill for region of fit
        app.add_electron_retarding_fit(
            voltage_KT_e_fit, current_KT_e_fit, solver.results["KT_e"]["value"])
        app.add_electron_retarding_fill(*KT_e_fit.poly.domain)

        # Electron Saturation fit and fill for region of fit
        app.add_electron_saturation_fit(
            voltage_I_es_fit, current_I_es_fit, solver.results["I_es"]["value"])
        app.add_electron_saturation_fill(*I_es_fit.poly.domain)

        # Ion Saturation fit and fill for region of fit or domain area
        app.add_ion_saturation_fit(
            voltage_I_is_fit, current_I_is_fit, solver.results["I_is"]["value"],
        )
        app.add_ion_saturation_fill(*I_is_fit.poly.domain)

        # turn on legend
        app.legend()

        # add 1der plot
        app.add_1derivative(solver)
    def make_plot_basic(self, solver):
        app = self
        # make xdomain for fit plotting
        xdomain = [np.min(solver.data.voltage), np.max(solver.data.voltage)]

        # Make Plots ###
        # add probe trace
        app.add_probe_trace(solver.data.voltage, solver.data.current)
        # add electron current
        app.add_semilogy_trace(solver.data.voltage, solver.data.current)
        # mark deleted points
        app.add_deleted_points(solver.deleted_data.voltage,
                               solver.deleted_data.current)

        # Add to Plasma Properties to Plots
        #app.add_floating_potential(solver.results["V_f"]["value"])
        #app.add_plasma_potential(solver.results["V_s"]["value"])

        # turn on legend
        app.legend()
        # add 1der plot
        app.add_1derivative(solver)
    
    def clear_plots(self):
        self.ax1.cla()
        self.ax2.cla()
        self.ax3.cla()
    
    def save(self):
        super().save()
        self.frame3.save()




def other():
    # Make save all button
    # btn_save_all = tk.Button(master=self, text="Save Both and Next",
    #                         command=lambda: save_figures(fig, saveas, root))
    # btn_save_all.pack(side=tk.BOTTOM, expand=1, fill=tk.BOTH)

    if None:
        canvas.mpl_connect(
            "key_press_event", lambda event: print(f"you pressed {event.key}"))
        canvas.mpl_connect("key_press_event", key_press_handler)

        # Make quit button
        button_quit = tk.Button(master=root, text="Quit", command=root.destroy)
        button_quit.pack(side=tk.BOTTOM, expand=1, fill=tk.BOTH)




from pstl.utls import constants as c
from pstl.utls.objects import setup as object_setup

available_plasma_classes = {
    0   :   ["custom"],
    1   :   ["xenon"],
    2   :   ["krypton"],
    3   :   ["argon"],
    4   :   ["neon"],
}

def setup(settings, *args, **kwargs):
    """
    Creates and returns a plasma object based on settings dictionary passed in.
    The settings parameter must have keys 'neutral_gas' and 'masses'. 'masses'
    must be a dictionary where at least 'm_i' must be defined, 'm_e' may be ommited
    in a custom plasma case. If a known neutral_gas type is choosen such as Xenon, 
    then 'm_i' may be ommitted as well.
    
    Keys:
        'netural_gas'   : str   ->  name of plasma['cylinderical', 'spherical', or 'planar']
        'masses'        : dict  ->  masses of ions and electrons in kg    [{'m_i:<value>, (optional) 'm_e':<value>}]
    (optional)
        'name'      : str   ->  name designation for plasma object
        'amu'       : bool  ->  if True, then m_i and m_e are given in amu instead of kg
        'args'      : tuple ->  addional position arguments
        'kwargs'    : dict  ->  addional keyword arguments

    Returns: Probe Object

    Other Notes:
        mass of an electron is 1/1836 in amu --or-- 5.4466e-4
        """
    shape = settings.pop("neutral_gas").lower()
    if shape in available_plasma_classes[0]:
        Plasma_ = Plasma
    elif shape in available_plasma_classes[1]:
        Plasma_ = XenonPlasma
    elif shape in available_plasma_classes[2]:
        Plasma_ = KryptonPlasma
    elif shape in available_plasma_classes[3]:
        Plasma_ = ArgonPlasma
    elif shape in available_plasma_classes[4]:
        Plasma_ = NeonPlasma
    else:
        raise ValueError("'%s' is not a valid option."%(shape))
    masses = settings["masses"]
    amu = settings.get("amu", False)
    if amu is True:
        settings["m_i"] = (masses["m_i"]*c.m_p if masses["m_i"] is not None else None) if "m_i" in masses else None
        settings["m_e"] = (masses["m_e"]*c.m_p if masses["m_e"] is not None else None) if "m_e" in masses else None
    output: Plasma = object_setup(
        *args,
        settings=settings,
        builders={},
        Builder=Plasma_,
        **kwargs,
    )
    return output


class Plasma:
    def __init__(self, m_i, m_e=c.m_e, neutral_gas=None, name=None,description=None,*args, **kwargs) -> None:
        self._m_i = m_i
        self._m_e = m_e
        self._neutral_gas = neutral_gas
        self._name = name
        self._description = description

    @property
    def m_i(self):
        return self._m_i

    @property
    def m_e(self):
        return self._m_e

    @property
    def neutral_gas(self):
        return self._neutral_gas

    @property
    def name(self):
        return self._name
    @name.setter
    def name(self, string):
        if isinstance(string, (str, type(None))):
            self._name = string
        else:
            raise TypeError("'%s' Must be a str or None type, not %s"%(str(string),str(type(string))))
    @property
    def description(self):
        return self._description
    @description.setter
    def description(self, string):
        if isinstance(string, (str, type(None))):
            self._description = string
        else:
            raise TypeError("Description change must be str or None type, not type '%s'"%(str(type(string))))


class XenonPlasma(Plasma):
    def __init__(self, m_i=None, m_e=c.m_e, name=None,*args, **kwargs) -> None:
        neutral_gas = "Xenon"
        if m_i is None:
            m_i = 131.29*c.m_p  # amu*kg -> kg
        super().__init__(m_i, m_e, neutral_gas, name, *args, **kwargs)


class ArgonPlasma(Plasma):
    def __init__(self, m_i=None, m_e=c.m_e, name=None, *args, **kwargs) -> None:
        neutral_gas = "Argon"
        if m_i is None:
            m_i = 39.948*c.m_p  # amu*kg -> kg
        super().__init__(m_i, m_e, neutral_gas, name, *args, **kwargs)

class NeonPlasma(Plasma):
    def __init__(self, m_i=None, m_e=c.m_e, name=None, *args, **kwargs) -> None:
        neutral_gas = "Neon"
        if m_i is None:
            m_i = 20.1797*c.m_p  # amu*kg -> kg
        super().__init__(m_i, m_e, neutral_gas, name, *args, **kwargs)

class KryptonPlasma(Plasma):
    def __init__(self, m_i=None, m_e=c.m_e, name=None, *args, **kwargs) -> None:
        neutral_gas = "Krypton"
        if m_i is None:
            m_i = 283.798*c.m_p  # amu*kg -> kg
        super().__init__(m_i, m_e, neutral_gas, name, *args, **kwargs)


import numpy as np

from pstl.utls import constants as c
from pstl.diagnostics.probes.langmuir.single.analysis.ion_current.thin.shape_func_helpers import _function

# Functions for calculating current
def cylinderical_ab_calculator(r_p, lambda_De):
    ratio = r_p/lambda_De
    a = 1.18 - 0.00080*np.power(ratio, 1.35)
    b = 0.0684 + np.power(0.722+0.928*ratio, -0.729)
    return a, b

def cylinderical_function(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_De):
    a,b = cylinderical_ab_calculator(r_p, lambda_De)
    return _partial_function(voltage, area, n0, V_s, m_i, KT_e, a, b)

def spherical_ab_calculator(r_p,lambda_De):
    ratio = r_p/lambda_De
    a = 1.58 + np.power(-0.056 + 0.816*ratio, -0.744)
    b = -0.933 + np.power(0.0148+0.119*ratio, -0.125)
    return a, b

def spherical_function(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_De):
    a, b =spherical_ab_calculator(r_p, lambda_De)
    return _partial_function(voltage, area, n0, V_s, m_i, KT_e, a, b)

def planar_ab_calculator(r_p,lambda_De):
    ratio = r_p/lambda_De
    a = np.exp(-0.5)*np.sqrt(2*np.pi)*(2.28*np.power(ratio, -0.749))
    b = 0.806*np.power(ratio, -0.0692)
    return a, b

def planar_function(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_De):
    a, b = planar_ab_calculator(r_p, lambda_De)
    partial = _partial_function(voltage, area, n0, V_s, m_i, KT_e, a, b)
    standard = _function(area, n0, m_i, KT_e)
    return np.add(partial, standard)


# Functions for calculating current (position arguments important for wrappers here) #######
def wrapper_cylinderical_function(voltage, V_s, KT_e, area, n0, m_i, r_p, lambda_De):
    return cylinderical_function(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_De)


def wrapper_spherical_function(voltage, V_s, KT_e, area, n0, m_i, r_p, lambda_De):
    return spherical_function(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_De)


def wrapper_planar_function(voltage, V_s, KT_e, area, n0, m_i, r_p, lambda_De):
    return planar_function(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_De)
#############################################################################################


# generalized
def _partial_function(voltage, area, n0, V_s, m_i, KT_e, a, b):
    A = (c.e*n0*area)*np.sqrt(c.e*KT_e/(2*np.pi*m_i))*a

    x = np.divide(V_s-voltage, KT_e)

    with np.errstate(invalid="ignore"):
        power_x = np.power(x, b)

    current = np.multiply(
        A,
        power_x
    )

    return current



from . import (
    thin,
    transitional,
    thick,
)
from .helpers import check_for_ion_current_fit
from .methods import (
    available_functions,
    available_methods,
    options,
)
from .methods import (
    get_ion_current,find_ion_current
)
from .methods import (
    thin_method, thin_func_method, transitional_method, thick_method,
)
from .methods_helpers import (default_fit_kwargs)

def info():
    msg = """
        rp: Radius of Probe
        LDe: Electron Debye Length

        Let rp/LDe = ratio

        Then
        If ratio <= 3       ->  Thick Sheath
        If 3 < ratio < 50   ->  Transitional Sheath
        If ratio >= 50      ->  Thin Sheath
        """
    print(msg)


from typing import Dict, Any

import numpy as np

from pstl.utls import constants as c

from .algorithm_helpers import(
    get_plasma_property_sorter,
)
from .algorithm_helpers import(
    default_methods,
    default_plasma_properties_to_get,
)

from ..ion_current import get_ion_current

def _topham(voltage, current, area, m_i, m_e=c.m_e,
            methods={},
            smooth=False, itmax=1, convergence_percent=1,
            *args, **kwargs) -> Dict:
    properties = None
    if not isinstance(methods, dict):
        raise ValueError(
            "'methods' must be a dictionary not: ", type(methods))
    # convert convergence percent to a decimal
    convergence_decimal = convergence_percent/100

    # overwrite methods if passed in
    methods_to_use = dict(default_methods)
    methods_to_use.update(methods)

    # overwrite properties if passed in (not implemented yet)
    if properties is None:
        properties_to_get = list(default_plasma_properties_to_get)
    elif isinstance(properties, list):
        properties_to_get = list(properties)
    else:
        raise ValueError(
            "'properteies' must be a list or None not: ", type(methods))

    # set up results dictionary for returns
    results = {}
    for plasma_property in properties_to_get:
        results[plasma_property] = {'value': None, 'other': None}

    # see what kwargs are given and creat a new dictionary
    func_kwargs = {}
    for key in methods_to_use.keys():
        func_kwargs[key] = kwargs.get(key+"_kwargs", {})

    # Zero Step (optional):
    # smooth data for find better fit region

    # First Step:
    # Get Floating Potential
    # -> V_f
    key = "V_f"
    results[key]["value"], results[key]["other"] = get_plasma_property_sorter(
        key, methods_to_use[key], *args, **func_kwargs[key])

    # Intermediate Step (optional for convergence speedup):
    # Do Either @ V_bias << V_f:
    # 1) Fit a linear line in Ion Saturation Region
    # 2) Take the resonable minimum value of the Ion Saturation Region
    # 3) Fit a power term (i.e. 1/2) to the Ion Saturation Region
    # Then either use the fits to subtract out ion current from the total probe current
    # to get a approximate electron only current, or a flat value across the board of the
    # minimum ion current (may lead to artifcial bi-Maxwellian)
    # Ie = Iprobe - Ii  --or-- Iprobe = Ie+Ii (note: Ii convention is negative)
    # -> temp I_i
    I_i, I_i_extras = get_ion_current(
        "spherical",
    )

    convergence = False
    it = 0
    vs_old = float("inf")
    key_convergence = "V_s"
    while not convergence and it < itmax:
        # number of iterations
        it += 1

        # Second Step:
        # Find Rough Exponential Fit after Floating Potential in the Electron Retarding Region
        # Note: Ii is removed this should be only Ie
        # -> KT_e

        # Third Step:
        # Find Electron Saturation Exponential Fit
        # @ V_bias >> V_s (plasma or space potential) before probe breakdown (plasma created due to accelerated elcectrons),
        # There should be no ion current.
        # Note: Theory says I_es =
        # -> I_es = I_e(@V_s) = exp(m*V_s+b)

        # Fourth Step:
        # Find Plasma Potential via intersection point of Electron Retarding and Saturation Regions
        # May also be done using first or second derivate to locate like in lobbia but requries smoothed data
        # and often inconclusive
        # -> V_s

        # Fifth Step:
        # Find Ion Saturation Current
        # Either:
        # 1) Thin Sheath:
        #   Linear Ion Saturation Fit @ V_bias << V_f that intersects the x-axis at a V_bias > V_s
        #   -> I_is = I_i(@V_s) = m*V_s + b
        # 2) Thick Sheath/Orbital Motion Limited (OML):
        #   Power 1/2 Fit in the Ion Saturation Region @ V_bias << V_f
        #   I_i^2 = alpha*V_bias + beta
        #       Where: alpha = -(q_e*area*ni)^2*(2.1*q_e/(pi^2*m_i))
        #               beta = (q_e*area*ni)^2*(2.1*q_e/(pi^2*m_i))*V_s
        #   Note: in theory I_es = -I_is*exp(0.5)*sqrt(m_i/(2*pi*m_e))
        # questionable below:
        #   -> I_is = -I_es*exp(-0.5)*sqrt(2*pi*m_e/m_i) --or-- I_is = -exp(-0.5)*area*q_e*n_i*sqrt(KT_e/m_i)

        # combines while loop steps into a few lines
        for key in ["KT_e", "I_es", "V_s", "I_is"]:
            results[key]["value"], results[key]["other"] = get_plasma_property_sorter(
                key, methods_to_use[key], *args, **func_kwargs[key])

        # Repeat till Convergence on V_s from Intermediate Step to Last Step
        # using Ion Saturation Current fit to correct orginal probe current data to electron only current
        # i.e. I_probe,orginal - I_i,new = I_e,next_iteration_data

        vs_new = results[key_convergence]['value']
        difference = vs_old-vs_new
        realtive_difference = difference/vs_old
        abs_rel_diff = np.abs(realtive_difference)

        convergence = True if abs_rel_diff <= convergence_decimal else False

    # Last Step:
    # Once convergence is made, get n_i and n_e (ion and electron densities) and J_es and J_is (electorn and ion saturation current densities)
    # Electrons:
    # -> n_e = I_es/(area*q_e)*sqrt(2*pi*m_e/KT_e)
    # -> J_es = I_es/area
    # Ions:
    # if thin sheath:
    # n_i = I_is/(area*q_e)*sqrt(m_i/KT_e)
    # if thick sheath:
    # n_i = ((alpha*pi^2*m_i)/(2*q_e^3*area^2))^(0.5)

    # Debye Length
    # -> lambda_De = sqrt(KT_e*epsilon_0/(n_e*q_e^2))

    # combines while loop steps into a few lines
    for key in ["n_e", "n_i", "J_es", "J_is", "lambda_De"]:
        results[key]["value"], results[key]["other"] = get_plasma_property_sorter(
            key, methods_to_use[key], *args, **func_kwargs[key])

    return results


from abc import ABC, abstractmethod, abstractproperty

class PSTLObject(ABC):
    def __init__(self, name=None, description=None,*args, **kwargs) -> None:
        super().__init__()

        self.name = name
        self.description = description

    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, string):
        if isinstance(string, (str, type(None))):
            self._name = string
        else:
            raise TypeError("'%s' Must be a str or None type, not %s"%(str(string),str(type(string))))

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, string):
        if isinstance(string, (str, type(None))):
            self._description = string
        else:
            raise TypeError("Description change must be str or None type, not type '%s'"%(str(type(string))))


from pstl.utls.helpers import method_selector, method_function_combiner

from .shape_fit_methods import (
    cylinderical_fit_method,
    spherical_fit_method,
    planar_fit_method,
)
from .shape_func_methods import (
    cylinderical_func_method,
    spherical_func_method,
    planar_func_method,
)


def shape_fit_selector(shape):
    # select function based on integer_method
    func = method_selector(shape, available_methods, available_fit_functions)
    return func


def shape_func_selector(shape):
    # select function based on integer_method
    func = method_selector(shape, available_methods, available_func_functions)
    return func

# Declare available methods
available_methods = {
    0: ['cylinderical', 'cylindrical'],
    1: ['spherical'],
    2: ['planar', 'planer'],
}
# declare correspondin functions for available_methods
available_fit_functions = {
    0: cylinderical_fit_method,
    1: spherical_fit_method,
    2: planar_fit_method,
}
# declare correspondin functions for available_methods
available_func_functions = {
    0: cylinderical_func_method,
    1: spherical_func_method,
    2: planar_func_method,
}

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
fit_options = method_function_combiner(
    available_methods, available_fit_functions)

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
func_options = method_function_combiner(
    available_methods, available_func_functions)


from typing import Any, Callable, Iterable
import pprint

from pstl.utls.json import load_build_and_get_parameters

def loop_key(settings, keys):
    if len(keys) != 0:
        output = loop_key(settings[keys[0]], keys.remove(keys[0]))
    else:
        output = settings
    return output

    
        


def extract_to_args(definer, settings):
    export = dict(definer)
    for key in definer:
        export[key] = loop_key(settings,definer[key])

    return export

def setup(
        settings:   dict, 
        builders:   dict[str, Any],
        Builder:    Any,
        *args, 
        to_args:    dict[str, list[str]] = {},
        **kwargs) -> Any:
    
    # get parameters by loading and/or building from file
    parameters = load_build_and_get_parameters(
        *args,
        settings=settings,
        builders=builders,
        **kwargs,
    )

    # using the to_args dictionary after everything is built
    extracted_to_args = extract_to_args(to_args, settings) if len(to_args) != 0 else {} 
    
    # creates object from parameters (sourced from settings)
    output: Any = Builder(
        *args, 
        **extracted_to_args,
        **parameters, 
        **kwargs,
    )

    return output




from pstl.utls.helpers import method_selector
from pstl.utls.helpers import method_function_combiner

from . import thin, transitional, thick
#from .helpers import available_functions, available_methods

def find_ion_current(shape, *args, method=None, **kwargs):
    """
    Determines and returns only a 1D array of ion current contribution for each point of an IV trace. 
    Similar to get_ion_current, but no extras: dictionary is returned, just the 1D array of ion currents."""
    value, _ = get_ion_current(shape, method=method, *args, **kwargs)
    return value


def get_ion_current(shape, *args, method=None, ** kwargs):
    """
    Determines and returns a 1D array of ion current contribution for each point of an IV trace along with a dictionary with extras information.
    A default method is assumed unless specified via keyword argument 'method' to be 'thin_method'. The method's function is then called with
    argument 'shape', *args, **kwargs passed to the function. All called functions are templated to return ion_current: 1D array, extras: dictionary.
    """
    # set default -> (thin)
    method = 0 if method is None else method

    # select function based on integer_method
    func = method_selector(method, available_methods, available_functions)

    return func(shape, *args, **kwargs)

def thin_method(shape, *args, **kwargs):
    """
    Deterimes and then runs for the thin sheath ion current using a fitted experical function to the data.
    """
    func = thin.shape_fit_selector(shape)
    return func(*args, shape=shape,**kwargs)


def thin_func_method(shape, *args, **kwargs):
    """
    Determines and then runs for the thin sheath ion current using a theorical function to determine data.
    """
    func = thin.shape_func_selector(shape)
    return func(*args, shape=shape,**kwargs)


def transitional_method(shape, *args, **kwargs):
    func = transitional.shape_selector(shape)
    return func(*args, shape=shape,**kwargs)


def thick_method(shape, *args, **kwargs):
    func = thick.shape_selector(shape)
    return func(*args, shape=shape,**kwargs)

# Declare available methods
available_methods = {
    0: ['thin'],
    1: ['transitional'],
    2: ['thick', 'OML'],
    3: ['thin-func']
}
# Declare correspondin functions for available_methods
available_functions = {
    0: thin_method,
    1: transitional_method,
    2: thick_method,
    3: thin_func_method,
}

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
options = method_function_combiner(available_methods, available_functions)



class FailedLangmuirAlgorithmConvergence(Exception):
    """Raised when the Single probe Langmuir algorithm cannot converge on a sheath type"""

    def __init__(self, msg="Failed to converge in Langmuir algorithm"):
        self.msg = msg
        Exception.__init__(self, self.msg)
class FailedLangmuirAlgorithm(Exception):
    """Raised when the Single probe Langmuir algorithm has a general error"""

    def __init__(self, msg="General Langmuir algorithm error"):
        self.msg = msg
        Exception.__init__(self, self.msg)

class Flagged(Exception):
    """
    Raises this error when something has been flagged so it can be handeled
    """
    def __init__(self, msg=""):
        self.msg = "FLAGGED: "+ msg
        Exception.__init__(self, self.msg)



import string
from typing import Any, Optional, Union, Tuple, Sequence, Callable
import traceback

import numpy as np
import numpy.typing as npt
from numpy.polynomial import Polynomial as P
from numpy.linalg import LinAlgError
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from scipy.signal import savgol_filter, find_peaks
from scipy.interpolate import interp1d
import pandas as pd

from pstl.utls import constants as c
from pstl.utls.errors import FitConvergenceError, MissingReturnError, FunctionFitError
from pstl.utls.verify import verify_iterables, verify_1D_array, verify_type


class FunctionFit:
    def __init__(
            self,
            func: Callable[[npt.ArrayLike, tuple[int | float, ...]], int | float],
            coef: Sequence,
            poly: P,
            domain: Sequence,
            stats: dict,
    ) -> None:

        self.func = func
        self.coef = coef
        self.poly = poly
        self.domain = domain
        self.stats = stats

    def __call__(self, x: npt.ArrayLike, *args: Any, **kwds: Any) -> Any:
        return self.func(x, *self.coef)

    def xrange(self, n: int | None = None, domain: Optional[Sequence] = None):
        if domain is None:
            domain = self.domain  # type: ignore
        if n is None:
            n = int((domain[1]-domain[0])/1e-3)

        return np.linspace(domain[0], domain[1], n)  # type: ignore

    def yrange(self, n: int | None = None, domain: Optional[Sequence] = None):
        if domain is None:
            domain = self.domain  # type: ignore
        if n is None:
            n = int((domain[1]-domain[0])/1e-3)

        # type: ignore
        return self.__call__(np.linspace(domain[0], domain[1], n))


class ExponentialFit(FunctionFit):
    def __init__(self, poly: P, domain: npt.ArrayLike, stats: dict = {}) -> None:
        func = _fit_exponential_func
        coef = poly.convert().coef
        super().__init__(func, coef, poly, domain, stats)  # type: ignore


class PolynomialFit(FunctionFit):
    def __init__(self, poly: P, domain: npt.ArrayLike, stats: dict = {}) -> None:
        func = _fit_polynomial_func
        coef = poly.convert().coef
        super().__init__(func, coef, poly, domain, stats)  # type: ignore


class PowerFit(FunctionFit):
    def __init__(self, power: int | float, poly: P, domain: Sequence, stats: dict = {}, polarity: int = 1) -> None:
        func = _fit_power_func
        coef = poly.convert().coef
        self.power = power
        self._polarity = polarity
        super().__init__(func, coef, poly, domain, stats)

    def __call__(self, x: npt.ArrayLike, *args: Any, **kwds: Any) -> Any:
        result = self.func(
            self.power,
            x,  # type: ignore
            *self.coef)*self._polarity
        return result


class CustomFit(FunctionFit):
    def __init__(self, func, coef, domain, fit_xrange, stats={}):
        poly = P([0, 1], domain=fit_xrange)
        super().__init__(func, coef, poly, domain, stats)  # type: ignore


def make_CustomFit(func, voltage, current, coefs):
    domain = [np.min(voltage), np.max(voltage)]
    # make fit function
    indexs = np.where(current != 0)[0]
    vmin = voltage[indexs[0]]
    vmax = voltage[indexs[-1]]
    fit_xrange = [vmin, vmax]

    fit = CustomFit(func, coefs, domain, fit_xrange)
    return fit


def safe_sqrt(x):
    mask = x < 0
    sqrt_x = np.sqrt(np.abs(x))
    sqrt_x[mask] = np.nan
    return sqrt_x


def method_not_found(method, available_methods):

    # make table string to print
    table = "\n".join([f"{k}\t{v}" for k, v in available_methods.items()])

    # raise error and make table
    raise ValueError(
        f"Matching method not found: {method}\nChoose from one of the available options:\n{table}")


def get_method_options(method, available_methods) -> tuple[int, str]:
    imethod: int | None = None
    vmethod: str | None = None
    if isinstance(method, int) and method in available_methods:
        # perform action using list of strings tied to the integer key
        imethod = method
        vmethod = available_methods[method][0]

    elif isinstance(method, str):
        # perform action using the string as the method name
        # check if the string matches any of the available_methods
        for key, value in available_methods.items():
            if method.lower() in value:
                imethod = key
                vmethod = value[0]
    else:
        imethod = None
        vmethod = None

    # make returns
    if imethod is None or vmethod is None:
        returns: tuple[int, str] = (0, "")
        # will raise error below
        method_not_found(method, available_methods)
    else:
        returns: tuple[int, str] = (imethod, vmethod)

    return returns


def method_selector(method, available_methods, available_functions):
    imethod, _ = get_method_options(method, available_methods)
    func = available_functions[imethod]
    return func


def method_function_combiner(dict1, dict2):
    if set(dict1.keys()) != set(dict2.keys()):
        raise ValueError("The two dictionaries do not have the same keys!")
    new_dict = {}

    for key in dict1.keys():
        new_dict[key] = (dict1[key], dict2[key])

    return new_dict


def probe_radius_to_debye_length(lambda_D, radius_probe):
    l = np.abs(lambda_D)  # Debye length [m]
    rp = np.abs(radius_probe)  # radius of probe [m]

    ratio = np.divide(rp, l)

    if ratio <= 3:
        sheath = "thick/OML"
    elif ratio >= 50:
        sheath = "thin"
    else:
        sheath = "transitional"

    other = {"sheath_type": sheath}
    return ratio, other

def interpolate(x,X,Y,method=None):
    # find nearest neighbors (no sorting is performed)
    lenX = len(X)
    k = 0
    while X[k] <= x :
        if X[k] != x:
            if k+1 == lenX:
                raise ValueError("'{x}' not in domain of X")
            elif k+1 > lenX:
                raise ValueError("Something really went wrong")
            if X[k+1] > x:
                break
            elif X[k+1] < x:
                k += 1
            elif X[k+1] == x:
                y = Y[k+1]
                method = None
                break
        else:
            # return the value
            y = Y[k]
            method = None
            break
    # Perform interpolation
    if method is None:
        pass
    elif method == 'linear':
        if x != 0:
            f = interp1d(X[k:k+2],Y[k:k+2] )  # type: ignore
            y: float = f(x)
        else:
            pass
    elif method == 'exponential':
        if x != 0:
            xp = np.log(Y)
            f = interp1d(xp, X)
            y: float = f(x)
        else:
            pass
    elif method == 'logarithmic':
        if x != 0:
            xp = np.log(X)
            f = interp1d(Y, xp)
            y: float = np.exp(f(x))
        else:
            pass
    else:
        raise ValueError("Invalid value for interpolate parameter.")
    return y


def find_intersection(coefs1, coefs2):
    m1, b1 = coefs1
    m2, b2 = coefs2
    # check if lines are parallel
    if m1 == m2:
        return None

    # calculate intersection point
    x_intersect = (b2 - b1) / (m1 - m2)
    y_intersect = m1 * x_intersect + b1

    return np.array([x_intersect, y_intersect])


def count_placeholders(fmt):
    count = 0
    L = string.Formatter().parse(fmt)
    for x in L:
        if x[1] is not None:
            count += 1
    return count


def get_rsquared(yexperimental, ypredicted):
    if len(yexperimental) != len(ypredicted):
        raise ValueError(
            "Vector 1 and Vector 2 arrays must have the same length.")

    corr_matrix = np.corrcoef(yexperimental, ypredicted)
    corr = corr_matrix[0, 1]
    rsq = corr**2

    return rsq


def determine_step(step: int | float, step_type: str, len_x: int) -> int:
    if step_type == 'index':
        istep = int(step)
    elif step_type == "percent":
        istep = int(step*len_x)
    elif step_type == "distance":
        raise NotImplementedError
    else:
        raise ValueError(
            "'step_type' does not match 'index' or 'percent': %s" % (str(step_type)))

    # make istep at least one
    if istep == 0:
        istep += 1
    return istep



def normalize(x: np.ndarray) -> np.ndarray:
    """
    Returns the normalized ndarray of input array

    Parameters
    ----------
    x : arraylike
        A 1D array to be normalized.

    Returns
    -------
    np.ndarray
        Normalized np.ndarray.
    """
    return (x-np.min(x))/(np.max(x)-np.min(x))


def residual(xdata: np.ndarray, xdata_mean: np.ndarray):
    """
    Returns the residual (difference) between the two arrays

    Parameters
    ----------
    xdata :  arraylike
        A 1D array of experimental data
    xdata_mean : arraylike
        A 1D array of mean of the experimental data

    Returns
    -------
    np.ndarray
        Residual
    """
    # verify same length


def preproces_data(xdata: np.ndarray) -> np.ndarray:

    # initialize and set default
    outdata = np.ndarray(np.size(xdata), dtype=None)
    return outdata


def _fit_linear_func(x, b, m):
    return m*x+b


def _fit_exponential_func(x, *args):
    total = _fit_polynomial_func(x, *args)
    return np.exp(total)


def _fit_power_func(power, x, *args, invalid="ignore"):
    total = _fit_polynomial_func(x, *args)
    # print(total) if np.any(np.isnan(total)) else None
    # print(power)
    if invalid == "ignore":
        with np.errstate(invalid="ignore"):
            res = np.power(total, power)
    else:
        res = np.power(total, power)
    # print(res) if np.any(np.isnan(res)) else None
    return res


def _fit_polynomial_func(x, *args):
    total = 0.0
    for k, arg in enumerate(args):
        total = np.add(
            total,
            np.multiply(
                arg,
                np.power(
                    x,
                    k
                )
            )
        )

    return total


def ideal_gas_law_pressure_to_density(P_gas, T_gas: int | float = 300):
    """
    Returns neutral gas density [Torr]

    Parameters:
    P_gas [Torr]
    T_gas (optional=300) [K]
    """

    P = P_gas
    P = P*101325/760            # unit: J/m^3 <- [Torr]*[101325 Pa/760 Torr]*[1 J/m^3/1 Pa] NOTE: 1 atm = 101325 Pa = 760 Torr
    N = P/(c.K_B*T_gas )        # unit: m^-3 <- [J/m^3]/([J/K]*[K])
    return N                    # unit: m^-3


import numpy as np

from pstl.utls import constants as c
from pstl.utls.decorators import where_function_else_zero
#from pstl.utls.helpers import make_CustomFit, find_fit
from pstl.utls.functionfit import make_CustomFit, find_fit
from pstl.diagnostics.probes.langmuir.single.analysis.floating_potential import (
    check_for_floating_potential, get_below_floating_potential,
)

from .shape_fit_helpers import default_fit_linear_kwargs, default_fit_power_kwargs

from .shape_helpers import (
    wrapper_cylinderical_function,
    wrapper_spherical_function,
    wrapper_planar_function,
    cylinderical_ab_calculator,
    spherical_ab_calculator,
    planar_ab_calculator,
)

from .shape_domains import (
    cylinderical_domain_condition,
    spherical_domain_condition,
    planar_domain_condition,
)


def cylinderical_method(voltage, current, *args, **kwargs):
    # check  and setup xdata,ydata based on floaing potential
    V_f = kwargs.pop("V_f", None)
    V_f = check_for_floating_potential(V_f, voltage, current, **kwargs)
    iend, xdata, ydata = get_below_floating_potential(V_f, voltage, current)

    # I_i function
    r_p = kwargs.pop("r_p", None)
    lambda_De = kwargs.pop("lambda_De", None)
    _,b = cylinderical_ab_calculator(r_p, lambda_De)
    fit_kwargs = dict(default_fit_power_kwargs)
    fit_kwargs["power"] = b
    fit_kwargs["domain_range"] = [min(voltage), V_f]
    fit_kwargs.update(kwargs.pop("fit_kwargs", {}))
    fit = find_fit(xdata, ydata, **fit_kwargs)
    # silence error when computing
    with np.errstate(invalid="ignore"):
        I_i = fit(voltage)

    # condition for I_i is negative
    condition = I_i <= 0

    # get I_i
    I_i = np.where(
        condition,
        I_i,
        0,
    )

    # make returns
    extras = {"fit": fit, "method": "thin", "shape": "cylinderical"}
    return I_i, extras
    
def cylinderical_func_method(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_De, *args, **kwargs):

    func = where_function_else_zero(
        wrapper_cylinderical_function, cylinderical_domain_condition)

    coefs = (V_s, KT_e, area, n0, m_i, r_p, lambda_De)
    I_i = func(voltage, *coefs)

    fit = make_CustomFit(func, voltage, I_i, coefs)

    # make returns
    extras = {"method": "transitional", "shape": "cylinderical", "fit": fit}
    return I_i, extras

def spherical_method(voltage, current, *args, **kwargs):
    # check  and setup xdata,ydata based on floaing potential
    V_f = kwargs.pop("V_f", None)
    V_f = check_for_floating_potential(V_f, voltage, current, **kwargs)
    iend, xdata, ydata = get_below_floating_potential(V_f, voltage, current)

    # I_i function
    r_p = kwargs.pop("r_p", None)
    lambda_De = kwargs.pop("lambda_De", None)
    _,b = spherical_ab_calculator(r_p, lambda_De)
    fit_kwargs = dict(default_fit_linear_kwargs)
    fit_kwargs["power"] = b
    fit_kwargs["domain_range"] = [min(voltage), V_f]
    fit_kwargs.update(kwargs.pop("fit_kwargs", {}))
    fit = find_fit(xdata, ydata, **fit_kwargs)
    # silence error when computing
    with np.errstate(invalid="ignore"):
        I_i = fit(voltage)

    # condition for I_i is negative
    condition = I_i <= 0

    # get I_i
    I_i = np.where(
        condition,
        I_i,
        0,
    )

    # make returns
    extras = {"fit": fit, "method": "thick", "shape": "spherical"}
    return I_i, extras


def spherical_func_method(voltage, area, n0, V_s, m_i, KT_e,  r_p, lambda_De, *args, **kwargs):
    func = where_function_else_zero(
        wrapper_spherical_function, spherical_domain_condition)

    coefs = (V_s, KT_e, area, n0, m_i, r_p, lambda_De)
    I_i = func(voltage, *coefs)

    fit = make_CustomFit(func, voltage, I_i, coefs)

    # make returns
    extras = {"method": "transitional", "shape": "spherical", "fit": fit}
    return I_i, extras

def planar_method(voltage, current, *args, **kwargs):

    # check  and setup xdata,ydata based on floaing potential
    V_f = kwargs.pop("V_f", None)
    V_f = check_for_floating_potential(V_f, voltage, current, **kwargs)
    iend, xdata, ydata = get_below_floating_potential(V_f, voltage, current)

    # I_i function
    r_p = kwargs.pop("r_p", None)
    lambda_De = kwargs.pop("lambda_De", None)
    _,b = spherical_ab_calculator(r_p, lambda_De)
    fit_kwargs = dict(default_fit_linear_kwargs)
    fit_kwargs["power"] = b
    fit_kwargs["domain_range"] = [min(voltage), V_f]
    fit_kwargs.update(kwargs.pop("fit_kwargs", {}))
    fit = find_fit(xdata, ydata, **fit_kwargs)
    # silence error when computing
    with np.errstate(invalid="ignore"):
        I_i = fit(voltage)

    # condition for I_i is negative
    condition = I_i <= 0

    # get I_i
    I_i = np.where(
        condition,
        I_i,
        0,
    )

    # make returns
    extras = {"fit": fit, "method": "thick", "shape": "spherical"}
    return I_i, extras


def planar_func_method(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_De, *args, **kwargs):
    func = where_function_else_zero(
        wrapper_planar_function, planar_domain_condition)

    coefs = (V_s, KT_e, area, n0, m_i, r_p, lambda_De)
    I_i = func(voltage, *coefs)

    fit = make_CustomFit(func, voltage, I_i, coefs)

    # make returns
    extras = {"method": "transitional", "shape": "planar", "fit": fit}
    return I_i, extras





import numpy as np

from pstl.utls import constants as c

def _xs(voltage, V_s, KT_e, lambda_D):
    A = lambda_D*np.sqrt(2)/3
    x = 2*(V_s-voltage)/KT_e
    with np.errstate(invalid="ignore"):
        power_x = np.power(x, 0.75)
    return np.multiply(A, power_x)


def _function(area, n0, m_i, KT_e):
    return np.exp(-0.5)*c.e*n0*area*np.sqrt(c.e*KT_e/m_i)


def _constant_function(voltage, current, *args, **kwargs):
    return np.ones_like(voltage)*_function(*args)



from abc import ABC, abstractmethod, abstractproperty

import numpy as np

from pstl.utls.objects import setup as object_setup
from pstl.diagnostics.probes.classes import Probe

available_probe_classes = [
    ["cylinderical", "cylindrical"],
    ["spherical"],
    ["planer", "planar"],
]

def setup(settings, *args, **kwargs):
    """
    Creates and returns a Langmuir probe object  based on settings dictionary passed in.
    The settings parameter must have keys 'shape' and 'dimensions'.
    
    Keys:
        'shape'     : str   ->  geometery shape of probe ['cylinderical', 'spherical', or 'planar']
        'dimensions': dict  ->  probe dimensions    [{'diameter':<value>,'length':<value>}]
    (optional)
        'name'      : str   ->  name designation for probe 
        'args'      : tuple ->  addional position arguments
        'kwargs'    : dict  ->  addional keyword arguments

    Returns: Probe Object
    """
    # determine shape of 
    shape = settings.pop("shape").lower()
    if shape in available_probe_classes[0]:
        Probe = CylindericalSingleProbeLangmuir
    elif shape in available_probe_classes[1]:
        Probe = SphericalSingleProbeLangmuir
    elif shape in available_probe_classes[2]:
        Probe = PlanarSingleProbeLangmuir
    else:
        raise ValueError("'%s' is not a valid option."%(shape))
    #def raise_missing_key_error(key):
    #    raise KeyError("'%s' is not a defined key but needs to be"%(key))
    # check if plasma, probe, and data are either given here or a part of solver_kwargs
    #key = "dimensions"
    #solver_kwargs = settings[key] if key in settings else raise_missing_key_error(key)
    #to_args = {
    #    "diameter"    :   ["dimensions", "diameter"],
    #    "length"     :   ["dimensions", "length"],
    #    }

    # create new object with parameters (arguments)
    output_object: SingleProbeLangmuir = object_setup(
        *args,
        settings=settings,
        builders={},
        Builder=Probe,
    #    to_args=
        **kwargs,
    )

    return output_object

class SingleProbeLangmuir(Probe, ABC):
    def __init__(self, diameter, *args, shape="Unknown", **kwargs) -> None:
        self._diameter = float(diameter)
        self._radius = diameter/2
        self._shape = shape

        self._area = self.calc_area(diameter, *args, **kwargs)

    @property
    def diameter(self):
        return self._diameter

    @property
    def radius(self):
        return self._radius

    @property
    def area(self):
        return self._area

    @property
    def shape(self):
        return self._shape

    @abstractmethod
    def calc_area(self, diameter, *args, **kwargs) -> float:
        pass


class CylindericalSingleProbeLangmuir(SingleProbeLangmuir):
    def __init__(self, diameter, length, *args, **kwargs) -> None:
        shape = "cylinderical"
        super().__init__(diameter, length, *args, shape=shape, **kwargs)

        self._length = float(length)

    @property
    def length(self):
        return self._length

    def calc_area(self, diameter, length, *args, **kwargs) -> float:
        super().calc_area(diameter, length, *args, **kwargs)
        return diameter*np.pi*(length + diameter/4)


class PlanarSingleProbeLangmuir(SingleProbeLangmuir):
    def __init__(self, diameter, *args, **kwargs) -> None:
        shape = "planar"
        super().__init__(diameter, *args, shape=shape, **kwargs)

    def calc_area(self, diameter, *args, **kwargs) -> float:
        super().calc_area(diameter, *args, **kwargs)
        return diameter*diameter*np.pi/4


class SphericalSingleProbeLangmuir(SingleProbeLangmuir):
    def __init__(self, diameter, length=None, *args, **kwargs) -> None:
        if length is None:
            length = diameter

        shape = "spherical"
        super().__init__(diameter, length, *args, shape=shape, **kwargs)

        self._length = float(length)

    @property
    def length(self):
        return self._length

    def calc_area(self, diameter, length, *args, **kwargs) -> float:
        super().calc_area(diameter, length, *args, **kwargs)
        height = length-diameter/2
        radius = diameter/2
        cap = 2*np.pi*(radius*height)
        hemisphere = 2*np.pi*radius*radius
        return cap+hemisphere



from .classes import SingleProbeLangmuir, SphericalSingleProbeLangmuir, CylindericalSingleProbeLangmuir, PlanarSingleProbeLangmuir
from .classes import available_probe_classes
from pstl.utls.verify import verify_type

available_plasma_properties = [
    "V_f",
    "V_s",
    "KT_e",
    "lambda_De",
    "n_e",
    "n_i",
    "I_es",
    "I_is",
    "J_es",
    "J_is",
    "r_p/lambda_De",
    "sheath",
]


def shape_selector(shape:str):

    verify_type(shape,str)

    shape = shape.lower()
    if shape in available_probe_classes[0]:  # cylinderical
        probe = CylindericalSingleProbeLangmuir


from abc import ABC, abstractmethod, abstractproperty
from typing import Any, Sequence
import tkinter as tk
import pprint
import traceback

import pandas as pd

from pstl.gui import WidgetOrganizer, Widgets
from pstl.gui.langmuir.canvas import LinearSemilogyDoubleCanvasSingleProbeLangmuir
from pstl.gui.langmuir.results import SingleProbeLangmuirResultsFrame
from pstl.gui.langmuir.panel import SingleProbeLangmuirPanelFrame

from pstl.utls.plasmas import setup as plasma_setup
from pstl.utls.data import setup as data_setup
from pstl.utls.objects import setup as object_setup
from pstl.utls.abc import PSTLObject
from pstl.utls import errors as err
from pstl.utls.errors.langmuir import FailedLangmuirAlgorithm

from pstl.diagnostics.probes.langmuir.single import setup as probe_setup
from pstl.diagnostics.probes.langmuir.single.analysis.solvers.solvers import SingleLangmuirProbeSolver, PlasmaProbeSolver
from pstl.diagnostics.probes.langmuir.single.analysis.solvers.solvers import setup as solver_setup


class CombinedDataFrame(tk.Frame, PSTLObject):
    def __init__(self, Solver, Canvas, Panel, *args, 
                 master=None, cnf=None, 
                 **kwargs):
        # get kwargs for individual
        canvas_kwargs = kwargs.pop("canvas_kwargs", {})
        panel_kwargs = kwargs.pop("panel_kwargs", {})
        solver_kwargs = kwargs.pop("solver_kwargs", {})
        super().__init__(master, cnf=cnf, *args, **kwargs)
        PSTLObject().__init__(*args, **kwargs)

        # set defaults for canvas
        canvas_kwargs.setdefault("width", 5)
        canvas_kwargs.setdefault("height", 4)

        # set defaults for panel

        # make plotting canvas
        self._canvas = Canvas(  # LinearSemilogyDoubleCanvasSingleProbeLangmuir(
            self, cnf, **canvas_kwargs)

        # make control panel
        self._panel = Panel(  # SingleProbeLangmuirResultsFrame(
            self, cnf, **panel_kwargs)

        # make solver
        self._solver = Solver(**solver_kwargs)

        # pack Away!!
        self.canvas.grid(row=0, column=1, sticky="NSWE")
        self.panel.grid(row=0, column=0, sticky="NWE")

    @property
    def canvas(self):
        return self._canvas

    @property
    def panel(self):
        return self._panel

    @property
    def solver(self):
        return self._solver
    
class CombinedDataFrame2(tk.Frame, PSTLObject):
    def __init__(
            self, 
            solver: PlasmaProbeSolver, 
            Canvas, 
            Panel, 
            *args, 
                 master=None, cnf=None, name=None,
                 **kwargs):
        # get kwargs for individual
        canvas_kwargs = kwargs.pop("canvas_kwargs", {})
        panel_kwargs = kwargs.pop("panel_kwargs", {})
        super().__init__(master, *args, cnf=cnf,  **kwargs)
        PSTLObject().__init__(*args,name=name,**kwargs)

        # set defaults for canvas
        canvas_kwargs.setdefault("width", 5)
        canvas_kwargs.setdefault("height", 4)

        # set defaults for panel

        # make plotting canvas
        self._canvas = Canvas( #LinearSemilogyDoubleCanvasSingleProbeLangmuir(
            self, cnf, **canvas_kwargs)

        # make control panel
        self._panel =  Panel( #SingleProbeLangmuirResultsFrame(
            master=self, cnf=cnf, **panel_kwargs)

        # make solver
        self._solver = solver

        # make name l
        if name is None:
            name = "None"
        else:
            name = str(name)
        name_lbl = tk.Label(master=self,cnf=cnf, text=name)
        self._name_label = name_lbl

        # pack Away!!
        self.canvas.grid(row=0, column=1, rowspan=2,sticky="NSWE")
        self.panel.grid(row=1, column=0, sticky="NWE")
        self.name_label.grid(row=0, column=0, sticky="NWE")
    

    @property
    def canvas(self):
        return self._canvas

    @property
    def panel(self):
        return self._panel

    @property
    def solver(self):
        return self._solver
    
    @property
    def name_label(self):
        return self._name_label

singe_langmuir_probe_solver_setup_builders = {
    "plasma"    :   plasma_setup,
    "probe"     :   probe_setup,
    "data"      :   data_setup,
}

def LSSLCD_setup(settings, *args, **kwargs):
    """
    Creates and returns a LinearSemilogySingleLangmuirCombinedData object based on settings dictionary passed in.
    The settings parameter must have keys 'plasma', 'probe', and 'data', where in
    each is a dictionary with either a key being 'define' or 'file_load'. If 'define',
    there is another dictionary that defines all need properties to make the object. In
    the 'file_load' case, the entry is a string that is the file path to load a json file
    that has all the needed properties to make the object.
    
    Keys (mandatory):
        'plasma': dict[dict | str]   ->  Plasma object defining properties or path to json file 
        'probe' : dict[dict | str]   ->  Probe object defining properties or path to json file
        'data'  : dict[dict | str]   ->  Data object defining properties or path to json file
        'solver_kwargs' : dict[dict | str]  ->  kwargs for solver class either
    (optional)
        'name'          : str   ->  name designation for object
        'description'   : str   ->  description of object
        'args'          : tuple ->  addional position arguments
        'kwargs'        : dict  ->  addional keyword arguments
    Returns: Solver Object
    """
    def raise_missing_key_error(key):
        raise KeyError("'%s' is not a defined key but needs to be"%(key))
    # check if plasma, probe, and data are either given here or a part of solver_kwargs
    key = "solver_kwargs"
    solver_kwargs = settings[key] if key in settings else raise_missing_key_error(key)
    to_args = {
        "plasma"    :   ["solver_kwargs", "plasma"],
        "probe"     :   ["solver_kwargs", "probe"],
        "data"      :   ["solver_kwargs", "data"]
        }
    
    # create new object with parameters (arguments)
    output_object:LinearSemilogySingleLangmuirCombinedData = object_setup(
        *args,
        settings=settings,
        builders=singe_langmuir_probe_solver_setup_builders,
        Builder=LinearSemilogySingleLangmuirCombinedData,
        to_args=to_args,
        **kwargs,
    )

    return output_object

solver_setup_builders = {
    "plasma"    :   plasma_setup,
    "probe"     :   probe_setup,
    "data"      :   data_setup,
    "solver"    :   solver_setup,
}
def LSSLCD2_setup(settings, *args, **kwargs):
    """
    Creates and returns a LinearSemilogySingleLangmuirCombinedData object based on settings dictionary passed in.
    The settings parameter must have keys 'plasma', 'probe', and 'data', where in
    each is a dictionary with either a key being 'define' or 'file_load'. If 'define',
    there is another dictionary that defines all need properties to make the object. In
    the 'file_load' case, the entry is a string that is the file path to load a json file
    that has all the needed properties to make the object.
    
    Keys (mandatory):
        'solver' : dict[str,dict | str]  ->  Path or dict for creating solver class either
        'canvas_kwargs' : dict[str,dict | str]  ->  kwargs for canvas creating either
        'panel_kwargs' : dict[str,dict | str]  ->  kwargs for solver class either
    (optional)
        'name'          : str   ->  name designation for object
        'description'   : str   ->  description of object
        'args'          : tuple ->  addional position arguments
        'kwargs'        : dict  ->  addional keyword arguments
    Returns:  Object
    """
    
    # create new object with parameters (arguments)
    output_object:LinearSemilogySingleLangmuirCombinedData2 = object_setup(
        *args,
        settings=settings,
        builders=solver_setup_builders,
        Builder=LinearSemilogySingleLangmuirCombinedData2,
        **kwargs,
    )

    return output_object

class LinearSemilogySingleLangmuirCombinedData(CombinedDataFrame):
    def __init__(self, plasma, probe, data: pd.DataFrame, *args, master=None, cnf={}, **kwargs):
        Canvas = LinearSemilogyDoubleCanvasSingleProbeLangmuir
        Panel = SingleProbeLangmuirResultsFrame
        Solver = SingleLangmuirProbeSolver
        kwargs.setdefault("solver_kwargs",{})
        kwargs["solver_kwargs"].update({'Plasma':plasma,'Probe':probe,'Data':data})
        super().__init__(Solver, Canvas, Panel,*args, master=master, cnf=cnf, **kwargs)
        # can call self.canvas, self.panel, self.property

        # solve Trace
        self.solve()

        # make save and exit button
        btn_save_n_close = tk.Button(self,cnf=cnf,command=self.save_n_close,text="Save and Close")
        # make exit button
        btn_close = tk.Button(self,cnf=cnf,command=self.close,text="Close")

        # pack away!
        btn_save_n_close.grid(row=2,column=0,columnspan=2,sticky="NSWE")
        btn_close.grid(row=3,column=0,columnspan=2,sticky="NSWE")

        # store
        self.widgets = Widgets()
        self.widgets.buttons["save_n_close"] = btn_save_n_close
        self.widgets.buttons["close"] = btn_close
    
    def solve(self,*args, **kwargs):
        # solve for data
        self.solver.solve(*args, **kwargs)

        # make plots of solved for plasma parameters
        self.canvas.make_plot(self.solver)

        # pass data to results panel
        self.panel.results = self.solver.results

    def solve_with_preprocessing(self,*args, **kwargs):
        # pre process data
        self.solver.preprocess()

        # solve for data
        self.solver.solve(*args, **kwargs)

        # make plots of solved for plasma parameters
        self.canvas.make_plot(self.solver)

        # pass data to results panel
        self.panel.results = self.solver.results

    def save_n_close(self):
        self.panel.save()
        self.canvas.save()
        self.close()

    def close(self):
        self.destroy()

class LinearSemilogySingleLangmuirCombinedData2(CombinedDataFrame2):
    def __init__(
            self, 
            solver: SingleLangmuirProbeSolver, 
            *args,
            master=None, cnf={}, name=None, preprocess=True, **kwargs):
        Canvas = LinearSemilogyDoubleCanvasSingleProbeLangmuir
        Panel = SingleProbeLangmuirPanelFrame
        super().__init__(solver, Canvas, Panel, *args, master=master,cnf=cnf, name=name,**kwargs)
        # can call self.canvas, self.panel, self.property

        # solve Trace
        self.solve(*args,preprocess=preprocess,**kwargs)

        # make save and exit button
        btn_save_n_close = tk.Button(self,cnf=cnf,command=self.save_n_close,text="Save and Close")
        # make exit button
        btn_close = tk.Button(self,cnf=cnf,command=self.close,text="Close")

        # pack away!
        btn_save_n_close.grid(row=2,column=0,columnspan=2,sticky="NSWE")
        btn_close.grid(row=3,column=0,columnspan=2,sticky="NSWE")

        # store
        self.widgets = Widgets()
        self.widgets.buttons["save_n_close"] = btn_save_n_close
        self.widgets.buttons["close"] = btn_close

        # set the update buttons on the panel.control to pass retrieved algorithm_kwargs and to resolve if pushed
        for key, control_frame in self.panel.controls.widgets.frames["controls"].items():
            if "update" in control_frame.widgets.buttons:
                control_frame.widgets.buttons["update"].configure(command=self.updated_solve)

    def solve(self, *args, preprocess=True, **kwargs):
        # pre process data
        self.solver.preprocess(delete=preprocess)

        self.solve_no_preprocessing(self, *args, **kwargs)
    
    def solve_no_preprocessing(self, *args, **kwargs):

        try:
            # solve for data
            self.solver.solve(*args, **kwargs)

            # make plots of solved for plasma parameters
            self.canvas.make_plot(self.solver)
        except (err.FunctionFitError, err.FitConvergenceError) as e:
            print("Failed in functionfit: langmuir.solve")
            print(e)
            traceback.print_exc()
            # make basic plot
            self.canvas.make_plot_basic(self.solver)
            #raise e

        except FailedLangmuirAlgorithm as F:
            print("Failed in algorithm: langmuir.solve")
            print(F)
            traceback.print_exc()
            # make basic plot
            self.canvas.make_plot_basic(self.solver)
            #raise F
        except Exception as e:
            print("General Fail: langmuir.solve")
            print(e)
            traceback.print_exc()
            # make basic plot
            try:
                self.canvas.make_plot_basic(self.solver)
            except Exception as e2:
                print("Could not make basic plot: langmuir.solve")
                print(e2)
                traceback.print_exc()
                raise e2
            #pass
        except:
            print("Catch all Fail: langmuir.solve")
            traceback.print_exc()
            # make basic plot
            self.canvas.make_plot_basic(self.solver)
            #raise

        # pass data to results panel
        self.panel.displays.results = self.solver.results

    def updated_solve(self, *args, **kwargs):

        print("Updating")
        # get what already exists
        algorithm_kwargs = kwargs.get("algorithm_kwargs", {})
        # update algorthim_kwargs is dict of i.e. {"V_f_kwargs: value, ...} with retrieved from entries
        algorithm_kwargs.update(self.panel.get_updated_algorithm_kwargs_from_control_panel())
        # reset into kwargs
        kwargs["algorithm_kwargs"] = algorithm_kwargs
        
        # before passing into resolve, clear canvas
        self.canvas.clear_plots()

        # pass in and resolve
        self.solve_no_preprocessing(*args, **kwargs)

    def save_n_close(self):
        self.panel.save()
        self.canvas.save()
        self.close()

    def close(self):
        self.destroy()


class MultipleDataFrame(tk.Frame):
    def __init__(self, num, DataFrame, master=None, cnf={}, *args, **kwargs):
        super().__init__(master, cnf=cnf, *args, **kwargs)

        self._pages = [None]*num

    def create_pages(self, num, DataFrame, *args, **kwargs):
        for k in range(num):
            self.pages[k] = DataFrame(*args, **kwargs)

    @property
    def pages(self):
        return self._pages

    def show_page(self, n):
        n -= 1

        self.pages[n].tkraise()



import numpy as np

# domain conditions
def cylinderical_domain_condition(voltage, V_s, KT_e):
    return _cylinderical_spherical_domain_condition(voltage, V_s, KT_e)


def spherical_domain_condition(voltage, V_s, KT_e):
    return _cylinderical_spherical_domain_condition(voltage, V_s, KT_e)

def planar_domain_condition(voltage, V_s, KT_e):
    condition = np.divide(
        np.subtract(V_s, voltage),  # [V]
        KT_e                         # [eV]
    )

    cond1 = 3 < condition
    cond2 = condition < 30
    return np.logical_and(cond1, cond2)  # 10 < rp/l < 45

# generalized domain
def _cylinderical_spherical_domain_condition(voltage, V_s, KT_e):
    condition = np.divide(
        np.subtract(V_s, voltage),  # [V]
        KT_e                         # [eV]
    )
    return condition > 1



import tkinter as tk
from typing import Any
from tkinter import Misc
from typing_extensions import Literal

import numpy as np

from pstl.gui import Widgets

from pstl.diagnostics.probes.langmuir.single.helpers import available_plasma_properties

class SingleProbeLangmuirResultsControlFrame(tk.Frame):
    def __init__(
            self, 
            controls:dict[str,Any]={}, 
            *args, 
            master=None, cnf={}, 
            **kwargs):
        control_frames_kwargs = kwargs.pop("control_frames_kwargs",{})
        super().__init__(master, cnf, *args, **kwargs)

        # add widgets organizer
        self.widgets = Widgets()
        
        # if controls are not given
        control_frames = {}
        if isinstance(controls, dict):
            for key in available_plasma_properties:
                # get individual control_frame kwargs for keyed value
                control_frame_key_kwargs = control_frames_kwargs.get(key, {})
                # set master and cnf
                control_frame_key_kwargs.setdefault("master", self)
                control_frame_key_kwargs.setdefault("cnf", cnf)
                # creates the control frame either the default or probided function to make a framif key not in controls else controls[key](cnf=cnf,**)e and then packs it onto this frame
                if key not in controls:
                    control_frame_key_kwargs.setdefault("txt", key)
                    control_frames[key] = self.create_default_frame(key, **control_frame_key_kwargs) 
                else:
                    # may need to verify that controls value element for key is a callable function that makes and returns a frame
                    control_frames[key] = controls[key](**control_frame_key_kwargs)
        else:
            raise ValueError(
                "'controls' must be dictionary: {}".format(type(controls)))
        
        # Once all control frames are made, they must be packed
        # all can be in the same spot as tk.raise will overide which is on top
        for key in control_frames:
            frame = control_frames[key]
            frame.grid(
                row=0,
                column=0,
                columnspan=5,
                sticky="NWSE",
            )

        # save function controls
        self._controls = controls

        # save control_frames  
        self.widgets.frames["controls"] = control_frames

        # save control buttons
        #self.widgets.buttons["save"] = frame_save
        #self.widgets.buttons["back"]
        #self.widgets.buttons["update"]
        #self.widgets.buttons["cancel"]

        # save display status label
        #self.widgets.labels["status"]

        #self.update_texts(results=self.results)

        # set size of columns (prob not needed for now)
        self.columnconfigure(0, weight=1, uniform="fred")
    
    def save(self):
        self.widgets.buttons["save"].save()

    @property
    def controls(self):
        return self._controls
    
    def get_all_controls_entries(self):
        """
        Returns dict of i.e. {"V_f_kwargs: value, ...}
        """
        returns = {}
        for key,frame in self.widgets.frames["controls"].items():
            key_kwargs = key + "_kwargs"
            kwargs_values = frame.get_entries()
            returns[key_kwargs] = kwargs_values
        return returns

    def create_default_frame(self, key, *args, cnf={},**kwargs):
#ResultControl(key=key, create_func=None,**control_frame_key_kwargs)     #
        if key == "V_f":
            control_frame = ResultControl(key=key,cnf=cnf,**kwargs)
        elif key == "V_s":
            control_frame = ResultControl(key=key,cnf=cnf,**kwargs)
        elif key == "I_is":
            #control_frame = ResultControl(key=key,cnf=cnf,**kwargs)
            control_frame = ControlFunctionFit(cnf=cnf,**kwargs)
        elif key == "n_i":
            control_frame = ResultControl(key=key,cnf=cnf,**kwargs)
        elif key == "I_es":
            control_frame = ControlFunctionFit(cnf=cnf,**kwargs)
        elif key == "KT_e":
            control_frame = ControlFunctionFit(cnf=cnf,**kwargs)
        elif key == "n_e":
            control_frame = ResultControl(key=key,cnf=cnf,**kwargs)
        elif key == "J_es":
            control_frame = ResultControl(key=key,cnf=cnf,**kwargs)
        elif key == "J_is":
            control_frame = ResultControl(key=key,cnf=cnf,**kwargs)
        elif key == "lambda_De":
            control_frame = ResultControl(key=key,cnf=cnf,**kwargs)
        elif key == "r_p/lambda_De":
            control_frame = ResultControl(key=key,cnf=cnf,**kwargs)
        elif key == "sheath":
            control_frame = ResultControl(key=key,cnf=cnf,**kwargs)
        else:
            control_frame = ResultControl(key=key,cnf=cnf,**kwargs)

        return control_frame

    def create_electron_temperature(self, cnf, *args, **kwargs):
        print("Electron Temperature")

    def create_electron_saturation(self, cnf, *args, **kwargs):
        print("Electron saturation")


class ResultControl(tk.Frame):
    """
    This creates a frame that houses the controls and buttons for a control panel
    that is raised when the button on the ResultOutputs frame has.
    """
    def __init__(
            self, 
            key,
            create_func=None,
            *args, 
            master=None, cnf={}, 
            **kwargs):
        lbl_kwargs = kwargs.pop("label_kwargs",{})
        btn_kwargs = kwargs.pop("btn_kwargs",{})
        txt = kwargs.pop("txt","")
        super().__init__(master, cnf, *args, **kwargs)

        # create widgets
        self.widgets = Widgets()

        # create frame and populate with either default or call to create_func 
        kwargs.setdefault("key",key)
        kwargs.setdefault("txt",txt)
        self.create_place_holder(cnf, *args, **kwargs) if create_func is None else create_func(*args,cnf=cnf,**kwargs)
        
    def create_place_holder(self, cnf, *args, **kwargs):
        key = kwargs.pop("key")
        # create label
        lbl = create_label_place_holder(self,cnf, *args, **kwargs,)
        
        # pack away!
        lbl.grid(row=0,column=0,sticky="NSWE")

        # save access
        self.widgets.labels["not-implemented"] = lbl

        # set column size 
        self.columnconfigure(0, weight=1, uniform="fred")
    
    def get_entries(self):
        """
        Returns a empty dictionary for Control.get_all_controls_enteries()
        """
        return {}

# module functions
def create_label_place_holder(master, cnf, txt="",*args,**kwargs):
    # set defaults
    text = "Not yet implemented\n"+txt
    kwargs.setdefault("text", text)

    # create place holder label
    lbl = tk.Label(
        master = master,
        cnf = cnf,
        *args,
        **kwargs,
    )
    return lbl

class ControlEntryLabelFrame(tk.Frame):
    def __init__(self, lbl_txt, *args, master=None, cnf={},**kwargs):
        label_kwargs = kwargs.pop("label_kwargs",{})
        entry_kwargs = kwargs.pop("entry_kwargs",{})
        super().__init__(master, cnf, *args,**kwargs)

        # add widgets organizer
        self.widgets = Widgets()

        # creat label
        lbl = tk.Label(self, cnf,text=lbl_txt,**label_kwargs)
        # packit
        lbl.grid(row=0,column=0,sticky="NSEW")
        # creat label
        ent = tk.Entry(self, cnf,**label_kwargs)
        # packit
        ent.grid(row=0,column=1,sticky="NSEW")

        # save them
        self.widgets.labels["label"] = lbl
        self.widgets.entries["entry"] = ent

        # create short handel
        self._label = lbl
        self._entry = ent

        # column weight
        self.columnconfigure(0, weight=1, uniform="fred")
        self.columnconfigure(1, weight=1, uniform="ted")

    @property
    def label(self):
        return self._label
    
    @property
    def entry(self):
        return self._entry
    
    def clear_entry(self):
        self.widgets.entries["entry"].delete(0,tk.END)
    
    def get_label_text(self) -> str | None:
        value = self.widgets.labels["label"]["text"]
        return value

class ControlEntryLabelFrameInt(ControlEntryLabelFrame):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
    def get_entry(self) -> int | None:
        value = self.widgets.entries["entry"].get()
        if value.lower() == "none" or value == "":
            value = None
        else:
            value = int(value)
        return value
    
class ControlEntryLabelFrameFloat(ControlEntryLabelFrame):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
    def get_entry(self) -> float| None:
        value = self.widgets.entries["entry"].get()
        if value.lower() == "none" or value == "":
            value = None
        else:
            value = float(value)
        return value

class ControlFunctionFit(tk.Frame):
    def __init__(
            self, 
            master: Misc | None = None, 
            cnf: dict[str, Any] | None = {},
            *args,
            **kwargs): 
        vstart_kwargs = kwargs.get("vstart_kwargs",{})
        vend_kwargs = kwargs.get("vend_kwargs",{})
        istart_kwargs = kwargs.get("istart_kwargs",{})
        iend_kwargs = kwargs.get("iend_kwargs",{})
        fitmax_kwargs = kwargs.get("fitmax_kwargs",{})
        bitmax_kwargs = kwargs.get("bitmax_kwargs",{})
        min_points_kwargs = kwargs.get("min_points_kwargs",{})
        threshold_rmse = kwargs.get("threshold_rmse_kwargs",{})
        update_kwargs = kwargs.get("update_kwargs",{})
        cancel_kwargs = kwargs.get("cancel_kwargs",{})
        txt = kwargs.pop("txt","")

        super().__init__(master, cnf, *args, **kwargs)

        # add widgets organizer
        self.widgets = Widgets()

        # create four ControlLabelEntryFrames
        vstart = ControlEntryLabelFrameFloat(lbl_txt="vstart",master=self,cnf=cnf,**vstart_kwargs)
        vend = ControlEntryLabelFrameFloat(lbl_txt="vend",master=self,cnf=cnf,**vend_kwargs)
        istart = ControlEntryLabelFrameInt(lbl_txt="istart",master=self,cnf=cnf,**istart_kwargs)
        iend = ControlEntryLabelFrameInt(lbl_txt="iend",master=self,cnf=cnf,**iend_kwargs)
        fitmax = ControlEntryLabelFrameInt(lbl_txt="fitmax",master=self,cnf=cnf,**fitmax_kwargs)
        bitmax = ControlEntryLabelFrameInt(lbl_txt="bitmax",master=self,cnf=cnf,**bitmax_kwargs)
        min_points = ControlEntryLabelFrameInt(lbl_txt="min_points",master=self,cnf=cnf,**bitmax_kwargs)
        # create update and clear
        update_btn = tk.Button(self,cnf,text="Update",**update_kwargs)
        cancel_kwargs.setdefault("command", self.clear_enteries)
        cancel_btn = tk.Button(self,cnf,text="Cancel",**cancel_kwargs)

        # pack away
        vstart.grid(row=0,column=0,sticky="NSEW")
        vend.grid(
            row=0,
            column=1,
            sticky="NSWE",
        )
        istart.grid(row=1,column=0,sticky="NSEW")
        iend.grid(
            row=1,
            column=1,
            sticky="NSWE",
        )
        fitmax.grid(
            row=2,
            column=0,
            sticky="NSWE",
        )
        bitmax.grid(
            row=2,
            column=1,
            sticky="NSWE",
        )
        min_points.grid(
            row=3,
            column=0,
            sticky="NSWE",
        )
        # pack away
        update_btn.grid(
            row=5,
            column=0,
            sticky="NSEW",
        )
        cancel_btn.grid(
            row=5,
            column=1,
            sticky="NSEW",
        )

        # save them
        self.widgets.frames["vstart"] = vstart
        self.widgets.frames["vend"] = vend
        self.widgets.frames["istart"] = istart
        self.widgets.frames["iend"] = iend
        self.widgets.frames["fitmax"] = fitmax
        self.widgets.frames["bitmax"] = bitmax
        self.widgets.frames["min_points"] = min_points

        self.widgets.buttons["update"] = update_btn
        self.widgets.buttons["cancel"] = cancel_btn

        # create short handel
        self._vstart = vstart
        self._vend = vend
        self._istart = istart
        self._iend = iend
        self._fitmax = fitmax
        self._bitmax = bitmax

        # column weight
        self.columnconfigure(0, weight=1, uniform="fred")
        self.columnconfigure(1, weight=1, uniform="fred")
        #self.rowconfigure(0, weight=1, uniform="fred")
        #self.rowconfigure(1, weight=1, uniform="fred")
    
    @property
    def vstart(self):
        return self._vstart
    
    @property
    def vend(self):
        return self._vend

    @property
    def istart(self):
        return self._istart
    
    @property
    def iend(self):
        return self._iend

    @property
    def fitmax(self):
        return self._fitmax
    
    @property
    def bitmax(self):
        return self._bitmax
    
    def clear_enteries(self):
        for key in self.widgets.frames:
            #self.widgets.frames[key].widgets.entries["entry"].delete(0,tk.END)
            self.widgets.frames[key].clear_entry()
    def get_entries(self):
        fit_kwargs = {}
        for key in self.widgets.frames:
            label_text = self.widgets.frames[key].get_label_text()
            entry_value = self.widgets.frames[key].get_entry()
            if entry_value is None:
                _ = fit_kwargs.pop(label_text,None)
            else:
                fit_kwargs[label_text] = entry_value
        returns = {"fit_kwargs": fit_kwargs}
        return returns




import tkinter as tk
import traceback
from PIL import Image, ImageTk
from matplotlib import style
import matplotlib.pyplot as plt

from pstl.extras.images import pstl_png_path, pstl_ico_path, pstl_16_ico_path, pstl_32_ico_path
from pstl.gui.langmuir import LSSLCD2_setup
from pstl.utls.errors.langmuir import Flagged


def gui_langmuir(settings:dict):
    style.use("bmh")
    plt.rcParams.update({"font.size": 10})

    # initiate app
    app = tk.Tk()
    app.title("PSTL GUI Langmuir")
    app.iconphoto(
        True, 
        ImageTk.PhotoImage(Image.open(pstl_32_ico_path)),
        ImageTk.PhotoImage(Image.open(pstl_16_ico_path))
    )
    
    # Try to run the solver
    try: 
        page = LSSLCD2_setup(settings, master=app)
        # pack it on
        page.pack()
    except Exception as e:
        file = settings["name"]
        print("\nFAILED in gui_langmuir: ",file,"\n")
        print(e)
        traceback.print_exc()
        #with open("/home/tyjoto/janus/temp/fail.tab") as f:
        #    f.write(file+"\n")
        pass
    except:
        file = settings["name"]
        print("\nFAILED in gui_langmuir: ",file,"\n")
        traceback.print_exc()
        #with open("/home/tyjoto/janus/temp/fail.tab") as f:
        #    f.write(file+"\n")
        pass
    else:
        def save_n_close():
            page.save_n_close()
            app.destroy()
        # save and close
        btn_savenclose = tk.Button(app,text="Save and Close app", command=save_n_close)
        btn_savenclose.pack()
    finally:
        def raise_flagged(app,msg=""):
            app.destroy()
            #print("FLAGGED: ",msg)
            raise Flagged(msg)
        # close button
        btn = tk.Button(app,text="Close App",command=app.destroy)
        btn.pack()

        # flag button
        flag_str = repr("{0}".format(str(settings.get("name","Unknown"))))
        btn_flag = tk.Button(app,text="Flag",command=lambda: raise_flagged(app, flag_str))
        btn_flag.pack()
        try:
            # run loop
            app.mainloop()
        except Flagged as e:
            app.destroy()
            raise Flagged



from .helpers import shape_selector
from .shape_methods import (
    cylinderical_method,
    spherical_method,
    planar_method,
)
from .helpers import (
    available_methods,
    available_functions,
    options,
)


from typing import Dict, Any
import traceback
import pprint

import numpy as np
import pandas as pd

from pstl.utls import constants as c
from pstl.utls.constants import lambda_D as get_lambda_D
from pstl.utls.decorators import absorb_extra_args_kwargs, add_empty_dict
get_lambda_D = add_empty_dict(absorb_extra_args_kwargs(get_lambda_D))
from pstl.utls.errors.langmuir import FailedLangmuirAlgorithmConvergence, FailedLangmuirAlgorithm
from pstl.utls.plasmas.sheath import get_probe_to_sheath_ratio

from .algorithm_helpers import(
    get_plasma_property_sorter,
    print_results,
    topham_configure,
    topham_get_ion_current_and_density,
    get_iteration_args_kwargs,
)
from .algorithm_helpers import(
    default_methods,
    default_plasma_properties_to_get,
    topham_get_ion_current_and_density
)

from ..ion_density import get_ion_density
from ..electron_density import get_electron_density
from ..plasma_potential import get_plasma_potential
from ..electron_saturation_current import get_electron_saturation_current, get_electron_saturation_current_density
from ..electron_temperaure import get_electron_temperature
from ..ion_saturation_current import get_ion_saturation_current, get_ion_saturation_current_density
from ..floating_potential import get_floating_potential
from ..ion_current import get_ion_current

def lobbia():
    return



######################################################################
######################################################################
######################################################################


def topham(voltage, current, shape, r_p, area, m_i, *args,
           m_e=c.m_e,
           methods={},
           smooth=False, itmax=9, convergence_percent=1,
           **kwargs) -> tuple[pd.DataFrame,Dict]:

    # initialize configuration
    (   convergence_decimal,
        methods_to_use,
        properties_to_get,
        results,
        func_args,
        func_kwargs,
    ) = topham_configure(convergence_percent, methods, *args, **kwargs)
    data = {"voltage": voltage,'current':current}
    rmse = float()
    V_f = float()
    V_f_extras = {}
    V_s = float()
    V_s_extras = {}
    n_e = float()
    n_e_extras = {}
    n_i = float()
    n_i_extras = {}
    KT_e = float()
    KT_e_extras = {}
    I_es = float()
    I_es_extras = {}
    I_is = float()
    I_is_extras = {}
    lambda_De = float()
    lambda_De_extras = {}
    ratio = float()
    ratio_extras = {}
    I_e = float()

    ionsat_method = 5



    try:


        # inialize sheath method
        sheath_method = "thin"

        # Zero Step (optional):
        # smooth data for find better fit region

        # First Step:
        # Get Floating Potential
        # -> V_f
        key = "V_f"
        #func_args[func_key], func_kwargs[func_key] = get_iteration_args_kwargs(
        #    iteration=1, key=func_key, args=args, kwargs=kwargs,
        #)
        V_f, V_f_extras = get_floating_potential(
            voltage, current, 
            *func_args[key],
            #method="consecutive", interpolate="linear",
            method=methods_to_use[key],
            **func_kwargs[key],
        )

        # Intermediate Step (optional for convergence speedup):
        # Do Either @ V_bias << V_f:
        # 1) Fit a linear line in Ion Saturation Region
        # 2) Take the resonable minimum value of the Ion Saturation Region
        # 3) Fit a power term (i.e. 1/2) to the Ion Saturation Region
        # Then either use the fits to subtract out ion current from the total probe current
        # to get a approximate electron only current, or a flat value across the board of the
        # minimum ion current (may lead to artifcial bi-Maxwellian)
        # Ie = Iprobe - Ii  --or-- Iprobe = Ie+Ii (note: Ii convention is negative)
        # -> temp I_i
        key = "I_is"
        I_i, I_i_extras = get_ion_current(
            shape,
            voltage, current,
            method=sheath_method, V_f=V_f,**func_kwargs[key],
        )
        I_i_fit = I_i_extras["fit"]


        # Initialize convergence and loop counters/exit-criteria
        convergence = False
        it = 0
        old = float("inf")

        # Start Convergence loop 
        # First step is thin sheath assumption, 
        # then continues with new sheath type (either transitional or thick (OML)) 
        # till within convergence percent using this iteration compared to last iteration
        # the last iteration values are sourced as the results, not the new as new is just used to show convergence
        # Thus if not thin sheath, the loop will run at for the new sheath type at least twice
        while not convergence and it < itmax:
            # number of iterations
            it += 1
            #print("\nIteration %i: (%s)"%(it,sheath_method))

        
            # Calculate electron current (I_e) from probe current (current) using determined ion current (I_i)
            I_e = np.subtract(current, I_i)

            # Second Step:
            # Find Rough Exponential Fit after Floating Potential in the Electron Retarding Region
            # Note: Ii is removed this should be only Ie
            # -> KT_e
            key = "KT_e"
            KT_e, KT_e_extras = get_electron_temperature(
                voltage, I_e, 
                *func_args[key],
                method="fit", V_f=V_f,
                **func_kwargs[key],
            )
            KT_e_poly = KT_e_extras["fit"].poly.convert()

            # Third Step:
            # Find Electron Saturation Exponential Fit
            # @ V_bias >> V_s (plasma or space potential) before probe breakdown (plasma created due to accelerated elcectrons),
            # There should be no ion current.
            # Note: Theory says I_es =
            # -> I_es = I_e(@V_s) = exp(m*V_s+b)
            key = "I_es"
            I_es, I_es_extras = get_electron_saturation_current(
                voltage, I_e, 
                *func_args[key],
                method="fit", elec_ret_poly=KT_e_poly, V_f=V_f,
                **func_kwargs[key],
            )
            I_es_poly = I_es_extras["fit"].poly.convert()

            # Fourth Step:
            # Find Plasma Potential via intersection point of Electron Retarding and Saturation Regions
            # May also be done using first or second derivate to locate like in lobbia but requries smoothed data
            # and often inconclusive
            # -> V_s
            key = "V_s"
            V_s, V_s_extras = get_plasma_potential(
                voltage, I_e, 
                *func_args[key],
                method="intersection", V_f=V_f, elec_ret_poly=KT_e_poly, elec_sat_poly=I_es_poly,
                **func_kwargs[key],
            )

            # Fifth Step:
            # Find Ion Saturation Current
            # Either:
            # 1) Thin Sheath:
            #   Linear Ion Saturation Fit @ V_bias << V_f that intersects the x-axis at a V_bias > V_s
            #   -> I_is = I_i(@V_s) = m*V_s + b
            # 2) Thick Sheath/Orbital Motion Limited (OML):
            #   Power 1/2 Fit in the Ion Saturation Region @ V_bias << V_f
            #   I_i^2 = alpha*V_bias + beta
            #       Where: alpha = -(q_e*area*ni)^2*(2.1*q_e/(pi^2*m_i))
            #               beta = (q_e*area*ni)^2*(2.1*q_e/(pi^2*m_i))*V_s
            #   Note: in theory I_es = -I_is*exp(0.5)*sqrt(m_i/(2*pi*m_e))
            # questionable below:
            #   -> I_is = -I_es*exp(-0.5)*sqrt(2*pi*m_e/m_i) --or-- I_is = -exp(-0.5)*area*q_e*n_i*sqrt(KT_e/m_i)
            # Last Step:
            # Once convergence is made, get n_i and n_e (ion and electron densities) and J_es and J_is (electorn and ion saturation current densities)
            # Electrons:
            # -> n_e = I_es/(area*q_e)*sqrt(2*pi*m_e/KT_e)
            # -> J_es = I_es/area
            # Ions:
            # if thin sheath:
            # n_i = I_is/(area*q_e)*sqrt(m_i/KT_e)
            # if thick sheath:
            # n_i = ((alpha*pi^2*m_i)/(2*q_e^3*area^2))^(0.5)

            # Debye Length
            # -> lambda_De = sqrt(KT_e*epsilon_0/(n_e*q_e^2))

            n_e, n_e_extras = get_electron_density(
                I_es, area=area, KT_e=KT_e, m_e=c.m_e, method="I_es",
            )
            #print("m_i\n", m_i)
            # get ionsaturation value at V_f based on V_s, I_es, n_e, KT_e, V_f (need V_s to solve)
            I_is, I_is_extras = get_ion_saturation_current(
                voltage, current, V_f=V_f, method=ionsat_method,
                I_i_fit=I_i_fit, I_i_method=sheath_method,
                V_s=V_s, n_e=n_e,T_e=KT_e, area=area,
                electron_sat=I_es, amu=c.kg_2_amu(m_i),
            )
            #print("I_is: ", I_is)
            # temp ion density
            key = "I_is"
            ion_current_kwargs = func_kwargs[key]
            n_i, n_i_extras = get_ion_density(
                I_is=I_is, voltage=voltage, current=current, area=area, KT_e=KT_e, m_i=m_i, method=sheath_method, shape=shape, r_p=r_p, lambda_De=lambda_De,**ion_current_kwargs,
            )
            # lambda_D
            lambda_De, lambda_De_extras = get_lambda_D(n_e, KT_e)
            lambda_Dei, lambda_Dei_extras = get_lambda_D(n_i, KT_e)

            # ratio
            ratio, ratio_extras = get_probe_to_sheath_ratio(r_p, lambda_De)
            # sheath_method = ratio_extras["sheath"]

            # Repeat till Convergence on V_s from Intermediate Step to Last Step
            # using Ion Saturation Current fit to correct orginal probe current data to electron only current
            # i.e. I_probe,orginal - I_i,new = I_e,next_iteration_data
            if ratio <= 3:
                sheath_method = "thick"
            elif ratio > 3 and ratio < 50:
                sheath_method = "transitional"
            elif ratio >= 50 and it == 1:
                #sheath_method = "thin"
                break
            elif ratio >= 50:
                errmsg = "Back to thin"
                raise ValueError(errmsg)
            else:
                errmsg = "WHa???!!!"
                raise ValueError(errmsg)

            if it == itmax and it==1:
                break
            elif it == itmax and it !=1:
                errmsg = "No convergence in Langmuir Topham algorithm was made after %i iterations"%(itmax)
                raise FailedLangmuirAlgorithmConvergence(errmsg)

            # get updated n_i
            # <<enter here>>
            key = "I_is"
            ion_current_kwargs = func_kwargs[key]
            I_i, I_i_extras, n_i, n_i_extras = topham_get_ion_current_and_density(
                voltage=voltage, current=current, shape=shape, sheath_method=sheath_method,
                area=area, m_i=m_i, KT_e=KT_e, r_p=r_p, lambda_De=lambda_De,
                V_f=V_f, V_s=V_s, n_e=n_e, n_i=n_i,ion_current_kwargs=ion_current_kwargs,ion_density_kwargs=ion_current_kwargs
            )

            # update I_is and I_is extras ## NEEEEEEEEDS ATTENETION
            I_i_fit = I_i_extras["fit"]
            # how to do ion saturation current for thick and transitional (i think below works)
            # get ionsaturation value at V_f based on V_s, I_es, n_e, KT_e, V_f (need V_s to solve)
            I_is, I_is_extras = get_ion_saturation_current(
                #voltage, current, V_f=V_f, method=ionsat_method,
                #I_i_fit=I_i_fit, I_i_method=sheath_method,
                #V_s=V_s, n_e=n_e,T_e=KT_e, area=area,
                voltage, current, V_f=V_f, method=ionsat_method,
                I_i_fit=I_i_fit, I_i_method=sheath_method,
                V_s=V_s, n_e=n_e,T_e=KT_e, area=area,
                electron_sat=I_es, amu=c.kg_2_amu(m_i),
            )



            results["V_f"]["value"] = V_f
            results["V_f"]["other"] = V_f_extras
            results["V_s"]["value"] = V_s
            results["V_s"]["other"] = V_s_extras
            results["I_is"]["value"] = I_is
            results["I_is"]["other"] = I_is_extras
            results["n_i"]["value"] = n_i
            results["n_i"]["other"] = n_i_extras
            results["I_es"]["value"] = I_es
            results["I_es"]["other"] = I_es_extras
            (results["J_es"]["value"], 
            results["J_es"]["other"]) = get_electron_saturation_current_density(area, I_es=I_es)
            (results["J_is"]["value"], 
            results["J_is"]["other"]) = get_ion_saturation_current_density(area, I_is=I_is)
            results["KT_e"]["value"] = KT_e
            results["KT_e"]["other"] = KT_e_extras
            results["n_e"]["value"] = n_e
            results["n_e"]["other"] = n_e_extras
            results["lambda_De"]["value"] = lambda_De
            results["lambda_De"]["other"] = lambda_De_extras
            #results["lambda_Dei"] = {"value":None, "other": None}
            #results["lambda_Dei"]["value"] = lambda_Dei
            #results["lambda_Dei"]["other"] = lambda_Dei_extras
            results["r_p/lambda_De"]["value"] = ratio
            results["r_p/lambda_De"]["other"] = ratio_extras
            results["sheath"]["value"] = ratio_extras['sheath']
            results["sheath"]["other"] = {}
            #print_results(results)

            new = lambda_De
            if it != 1:
                difference = old-new
                realtive_difference = difference/old
                rmse = np.sqrt(np.sum(np.power(realtive_difference, 2)))
            else:
                rmse = float("inf")
            old = new

            convergence = True if rmse <= convergence_decimal else False

        ##### REMOVE AFTER DEBUG ######

        #print("\nNumber of iterations: %i"%(it))
        #print("Convergence %f"%(rmse))
        ###############################

        results["V_f"]["value"] = V_f
        results["V_f"]["other"] = V_f_extras
        results["V_s"]["value"] = V_s
        results["V_s"]["other"] = V_s_extras
        results["I_is"]["value"] = I_is
        results["I_is"]["other"] = I_is_extras
        results["n_i"]["value"] = n_i
        results["n_i"]["other"] = n_i_extras
        results["I_es"]["value"] = I_es
        results["I_es"]["other"] = I_es_extras
        (results["J_es"]["value"], 
        results["J_es"]["other"]) = get_electron_saturation_current_density(area, I_es=I_es)
        (results["J_is"]["value"], 
        results["J_is"]["other"]) = get_ion_saturation_current_density(area, I_is=I_is)
        results["KT_e"]["value"] = KT_e
        results["KT_e"]["other"] = KT_e_extras
        results["n_e"]["value"] = n_e
        results["n_e"]["other"] = n_e_extras
        results["lambda_De"]["value"] = lambda_De
        results["lambda_De"]["other"] = lambda_De_extras
        #results["lambda_Dei"]["value"] = lambda_Dei
        #results["lambda_Dei"]["other"] = lambda_Dei_extras
        results["r_p/lambda_De"]["value"] = ratio
        results["r_p/lambda_De"]["other"] = ratio_extras
        results["sheath"]["value"] = ratio_extras['sheath']
        results["sheath"]["other"] = {}
        data = {'voltage': voltage, 'current': current,
                'current_e': I_e, 'current_i': I_i}
    except Exception as e:
        print("Fail: Topham Algorithm ")
        print(e)
        traceback.print_exc()
        raise FailedLangmuirAlgorithm
    except:
        print("Fail Catch all: Topham Algorithm ")
        traceback.print_exc()
        raise FailedLangmuirAlgorithm
    
    finally:
        data = pd.DataFrame(data)
        #print("DATA:\n")
        #with pd.option_context("display.max_rows",None,):
        #    print(data)
        #print()
        #print("RESULTS:")
        #print_results(results)
        return data, results
